// User interface for GenGeneric Scene Controller version 2.11
// Supports the following devices: 
// 		Evolve LCD1 5 button controller with LCD
//		Cooper RFWC5 5 button controller
//		Nexia One Touch 5 button controller with LCD
//		HomeSeer HS-WS100+ switch with 1 LED
//      HomeSeer HS-WD100+ dimmer with 7 LEDs
//		HomeSeer HS-WS200+ switch with 1 multi-color LED
//      HomeSeer HS-WD200+ dimmer with 7 multi-color LEDs
//      HomeSeer HS-FC200+ fan controller switch with 4 multi-color LEDs
//
// Copyright (C) 2020 Gustavo A Fernandez
// Thanks to Ron Luna and Ctrlable for contributing for RFWC5 support
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You can access the terms of this license at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html
//
var SID_SCENECONTROLLER   = "urn:gengen_mcv-org:serviceId:SceneController1"

var DEVTYPE_EVOLVELCD1     = "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD:1"
var DEVTYPE_COOPEREFWC5    = "urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5:1"
var DEVTYPE_NEXIAONETOUCH  = "urn:schemas-gengen_mcv-org:device:SceneControllerNexiaOneTouch:1"
var DEVTYPE_HOMESEERWS100  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS100:1"
var DEVTYPE_HOMESEERWD100  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD100:1"
var DEVTYPE_HOMESEERWS200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS200:1"
var DEVTYPE_HOMESEERWD200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD200:1"
var DEVTYPE_HOMESEERFC200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerFC200:1"

	// Evolve LCD1
var EVOLVELCD1 = {
	Id						: "EVOLVELCD1",
	Name                    : "Evolve LCD1",
	DefaultLcdVersion		: 39,
	HasScreen               : true,
	HasMultipleScreens      : true,
	ScreenWidth 		    : 65,
	RightJustifyScreenWidth : 67,
	HasPresetLanguages      : true,
	HasThremostatControl    : true,
    NumButtons              : 5,
    NumSceneEvents			: 0,
	MaxScroll	        	: 10,
    LastFixedSceneId        : 10,
	HasOffScenes            : true,
	HasIndicator			: true,
	HasMultiTap				: 0,
	HasLocalControl			: false,
	HasDirectSceneControl	: true,
	NumStatusLEDs			: 0,
	StatusLEDActionNames    : {"-": "Don't change LED Mode"}, 
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 30,
	DefaultScreen           : "C1",
	NumTemperatureScreens   : 3, // Preset Pages 8, 16, and 40
	StatusLEDActions		: "",
	StatusLEDTimeouts		: "",
	StatusLEDColors         : "",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1"],
								/* 1 - Scenes  */ []
							  ],
	MaxLabelSize			: 16,
	HasBattery				: false,
	HasButtonModes			: true,
	SendsSceneActivation	: true,
	SendsBasicSet			: true,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : false,
	CentralSceneSlowRefresh : false,
	DevType                 : DEVTYPE_EVOLVELCD1,
	DeviceXml               : "D_EvolveLCD1.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:EvolveLCD1",
	OldParamPrefix          : "LCD",

	GetDefaultModeString    : function(button, eventCode) { return "M"; },
	ScreenTypes 			: [
		{ prefix: "C", name: "Custum",      num: 6 },
		{ prefix: "T", name: "Temperature", num: 6 },
		{ prefix: "P", name: "Preset",      num: 41 }
		// { prefix: "W", name: "Welcome",     num: 1 },
	],
	CustomModeList 			: [
		"M", "T", "3", "4", "5", "6", "7", "8", "9", "X", "N", "P", "E"
	],
	CustomFontList          : [
		"Normal",
		"Compressed",
		"Inverted"
	/* - Screen MD V2 fonts. Not used by any of the supported scene controllets
		,
		"Compressed, Inverted",
		"No Scroll",
		"Compressed, No Scroll",
		"Inverted, No Scroll",
		"Compressed, Inverted, No Scroll"
	*/
	],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		var sceneBases = {
			C: 1,	// Custom      screen scenes start at base 1
			T: 31,	// Temperature screen scenes start at base C + 5 buttons * 6 custom screens
			W: 61,	// Welcome     screen scenes start at base T + 5 buttons * 6 temperature screens
			P: 66	// Preset      screen scenes start at base W + 5 buttons * 1 welcome screen
			// 271-300 - Unused	                 start at base P + 5 buttons * 41 preset screens
			// 301-330 - Lines 6-10 for Custom screens 1-6
			// 331-660 - Hold/release for above scenes (Central scene capable devices only)
			// 661-990 - Double Click for above scenes (Central scene V2 capable devices only)
			// 991-1000 - Unused
			// 1001-1030 - State 2 for multi-state custom screens. (1031-2000 - Unused)
			// ...
			// 8001-8030 - State 9 for multi-state custom screens. (8031 and above - Unused)
			// 9000, 9005, 9010, ... - Timeout scene for C1, C2, C3, etc.
		}

		// Screen timeout scen number starts at 9000 for C1, 9005 for C2, etc.
		if (timeout) {
			return sceneBases[screenType]+((screenNum-1)*5)+8999;
		}
		var base = sceneBases[screenType]
		if (!base) {
			base = 1
		}
		var buttonGroup = Math.floor((virtualButton-1)/5)
		return base+((screenNum-1)*5)+((virtualButton-1)%5)+(buttonGroup?(200+buttonGroup*100):0)+((state-1)*1000);
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		if (screenType == "C") {
			return true;
		}
		if (screenType == "T") {
			if (version >= 55) {
				return screenNum != 2; // Version 55 does not support temperature page 16
			}
			return screenNum != 3; // Older versions don't include temperature page 40
		}
		if (screenType != "P") {
			return false;
		}
		if (!SCObj.PresetScreens[screenNum]) {
			return false;
		}
		if (version <= 37) {
			return screenNum <= 18;
		}
		if (version <= 39) {
			return screenNum <= 26;
		}
		if (version <= 54) {
			return screenNum == 1 || screenNum == 8 || screenNum == 10 || screenNum == 13 || screenNum == 17 || (screenNum >= 19 && screenNum <= 26) || (screenNum >= 31 && screenNum <= 40);
		}
		/* Version 55 has one more screen than version 54 */
		    return screenNum == 1 || screenNum == 8 || screenNum == 10 || screenNum == 13 || screenNum == 17 || (screenNum >= 19 && screenNum <= 26) || (screenNum >= 31 && screenNum <= 41);
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		if (screenType == "T" && screenNum > 3) {
			return [1];	// Custom termperature screens don't have a language.
		}
		if (screenType == "P" || screenType == "T") {
			if (version <= 37) {
				return [1, 2, 3, 4, 5, 6, 7];
			}
			if (version <= 39) {
				return [1];
			}
			/* Version 55 has 1 more preset screen (41) than version 54 but preset 40 is Temperature and screen 41 does not support Chinese so this handles both cases.*/
			if (screenType == "P" && screenNum >= 37 && screenNum <= 39) {
				return [1, 3];
			}
			if (screenType == "T" && screenNum == 3) {
				return [1, 3];
			}
			return [1];
		}
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["All On","Low","All Off","Privacy Please","Service Room"],
		  /* 2  */ ["All On","Medium","Low","Night Light","All Off"],
		  /* 3  */ null,
		  /* 4  */ ["All On","Medium","Low","Mood","All Off"],
		  /* 5  */ ["All On","Medium","Low","Mood","All Off"],
		  /* 6  */ null,
		  /* 7  */ ["All On","Medium","Low","Night Light","All Off"],
		  /* 8  */ null,	/* Temperature */
		  /* 9  */ null,
		  /* 10 */ ["Drapery Open","Drapery Closed","Stop","Sheers Open","Sheers Closed"],
		  /* 11 */ ["Vanity","Shower","Night Light","All On","All Off"],
		  /* 12 */ ["Entry","Sconce","Bed Left","Bed Right","Good Night"],
		  /* 13 */ ["Vanity","Shower","Night Light","All On","All Off"],
		  /* 14 */ ["Entry","Kitchen","LivingRoom","BedRoom","MasterOff"],
		  /* 15 */ ["Welcome","Overhead","Bedroom","Privacy","All Off"],
		  /* 16 */ null, /* Temperature */
		  /* 17 */ null, /* Custom */
		  /* 18 */ ["Morning","Day","Evening","Night","Sleep"],
		  /* 19 */ ["All On","Living Room","Bed Room","Bath Room","All Off"],
		  /* 20 */ ["All On","Living Room","Bed Room","Low","All Off"],
		  /* 21 */ ["All On","Mood","All Off","Privacy Please","Service Room"],
		  /* 22 */ ["On/Off","Mood","Drapery","Reading","Night Light"],
		  /* 23 */ ["LR On/Off","LR Mood","BR On/Off","BR Mood","Drapery"],
		  /* 24 */ ["BO On/Off","BR Mood","LR On/Off","LR Mood","Drapery"],
		  /* 25 */ ["All On","All Off","Entry","Low","Mood"],
		  /* 26 */ ["All On","TurnDown","All Off","Service Please","Privacy Please"],
		  /* 27 */ null,
		  /* 28 */ null,
		  /* 29 */ null,
		  /* 30 */ null,
		  /* 31 */ ["All On","Entry","Bedside","Low","All Off"],
		  /* 32 */ ["Lighting","Drapery","Climate","Privacy Please","Service Room"],
		  /* 33 */ ["Lighting","Climate","Shades","Service Room","Privacy Please"],
		  /* 34 */ ["Lighting","Climate","Drapery","Master On","Master Off"],
		  /* 35 */ ["Lighting","Climate","Drapery","Bath On","Bath Off"],
		  /* 36 */ ["Lighting","Climate","Shading","Master On","Master Off"],
		  /* 37 */ ["All On","All Off","Privacy Please","Service Room","Language"],
		  /* 38 */ ["All On","Vanity","Toilet","Shower","All Off"],
		  /* 39 */ ["Sheers Open","Sheers Closed","Stop","Shade Open","Shade Closed"],
		  /* 40 */ null, /* temperature */
		  /* 41 */ ["Lighting","Climate","","All On","All Off"]
	],

	GetRowStyle				: function(button) {return ""},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return 99; },
	IndicatorHighlighting   : "Highlight line"
};

// Cooper RFWC5-Specific
var COOPERRFWC5 = {
	Id						: "COOPERRFWC5",
	Name                    : "Cooper RFWC5",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 5,
    NumSceneEvents			: 0,
	MaxScroll               : 5,
    LastFixedSceneId        : 5,
	HasOffScenes            : false,
	HasIndicator			: true,
	HasMultiTap				: 0,
	HasLocalControl			: false,
	HasDirectSceneControl	: true,
	NumStatusLEDs			: 0,
    StatusLEDActionNames    : {"-": "Don't change LED Mode"}, 
	HasCooperConfiguration  : true,
	MaxDirectAssociations   : 5,
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "",
	StatusLEDTimeouts		: "",
	StatusLEDColors         : "",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1"],
								/* 1 - Scenes  */ []
							  ],
	MaxLabelSize			: 0,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: true,
	SendsBasicSet			: true,
	SendsSwitchMultilevel   : true,
	SendsCentralScene       : false,
	CentralSceneSlowRefresh : false,
	DevType                 : DEVTYPE_COOPEREFWC5,
	DeviceXml               : "D_CooperRFWC5.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:CooperRFWC5",
	OldParamPrefix          : "RFWC5",

	GetDefaultModeString    : function(button, eventCode) {return "T";},
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"T", "M", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton+((state-1)*1000);
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Button 1","Button 2","Button 3","Button 4","Button 5"]
	],
	GetRowStyle				: function(button) {return ""},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};

// Nexia One-Touch-Specific
var NEXIAONETOUCH = {
	Id						: "NEXIAONETOUCH",
	Name                    : "Nexia One Touch",
	DefaultLcdVersion       : 1,
	HasScreen               : true,
	HasMultipleScreens      : false,
	ScreenWidth 		    : 65,
	RightJustifyScreenWidth : 67,
	HasPresetLanguages      : false,
	HasThremostatControl    : true,
    NumButtons              : 15,
    NumSceneEvents			: 0,
	MaxScroll               : 15,
    LastFixedSceneId        : 46,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 0,
	HasLocalControl			: false,
	HasDirectSceneControl	: true,
	NumStatusLEDs			: 0,
	StatusLEDActionNames    : {"-": "Don't change LED Mode"}, 
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 2,
	DefaultScreen           : "C1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "",
	StatusLEDTimeouts		: "",
	StatusLEDColors         : "",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1"],
								/* 1 - Scenes  */ []
							  ],
	MaxLabelSize			: 17,
	HasBattery				: true,
	HasButtonModes			: true,
	SendsSceneActivation	: true,
	SendsBasicSet			: true,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : false,
	DevType                 : DEVTYPE_NEXIAONETOUCH,
	DeviceXml               : "D_NexiaOneTouch.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:NexiaOneTouch",
	OldParamPrefix          : "NEXIA",

	GetDefaultModeString    : function(button, eventCode) { return "M";},
	ScreenTypes 			: [
		{ prefix: "C", name: "Custum",      num: 1 }
	],
	CustomModeList 			: [
		"M", "T", "H", "C"
	],
	CustomFontList          : [
		"Normal",
		"Large",
		"Inverted"
	/* - Screen MD V2 fonts. Not used by any of the supported scene controllets
		,
		"Large, Inverted",
		"No Scroll",
		"Large, No Scroll",
		"Inverted, No Scroll",
		"Large, Inverted, No Scroll"
	*/
	],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 330 : 0) + (taps-1)* 660 + (state-1) * 1000
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "C";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [ ],
	GetRowStyle				: function(button) {return ""},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return 99; },
	IndicatorHighlighting   : ""
};

// HomeSeer common functions

HomeSeer_GetDefaultModeString = function(button, eventCode) {
	var state = Math.floor(button/1000) + 1;
	button = button % 1000; 
	switch(button) {
		case 1:
		case 2:
			var color = "W"
			var action = "N"
			if (eventCode >= "2" && eventCode <= "5") {
				color = "WCMYRG".charAt(eventCode - 1);
				action = "L";
			} 
			return (button == 1 ? "OL" : "~OL") + action + color;
		default:
			// A:   Flashing red
			// B:   Restore
			// C-H: solid white, red, green, blue, cyan, magenta
			var color = "rB005"
			if (eventCode > "A" && eventCode <= "H") {
				color = "r-WRGBCM".charAt(eventCode.charCodeAt(0) - "A".charCodeAt(0))
			}
			return "-" + (eventCode == "B" ? "LR" : "LA") + color;
	}
}
// HomeSeer HS-WS100+ switch-specific
var HOMESEERWS100 = {
	Id						: "HOMESEERWS100",
	Name                    : "HomeSeer HS-WS100+ switch",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 2,
    NumSceneEvents			: 0,
	MaxScroll               : 2,
    LastFixedSceneId        : 2,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 3,
	HasLocalControl			: true,
	HasDirectSceneControl	: false,
	NumStatusLEDs			: 0,
	StatusLEDActionNames    : {"-": "Don't change LED Mode"}, 
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 5,
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "",
	StatusLEDTimeouts 		: "",
	StatusLEDColors         : "",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1", "2", "3"],
								/* 1 - Scenes  */ []
							  ],
	MaxLabelSize			: 1,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: false,
	SendsBasicSet			: false,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : true,
	DevType                 : DEVTYPE_HOMESEERWS100,
	DeviceXml               : "D_HomeSeerWS100.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:HomeSeerWS100",
	OldParamPrefix          : "HS",

	GetDefaultModeString    : HomeSeer_GetDefaultModeString,
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"O", "T", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 2 : 0) + (taps-1)*4 + (state-1) * 1000;
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Top button","Bottom button"]
	],
	GetRowStyle				: function(button) {return 'style="background-color:'+(button==1?'#F4F4F4':button==2?"#E8E8E8":"#DDFFDD")+';"';},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return labelIndex == 2 ? 0 : 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};

// HomeSeer HS-WD200+ dimmer-Specific
var HOMESEERWD100 = {
	Id						: "HOMESEERWD100",
	Name                    : "HomeSeer HS-WD100+ dimmer",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 2,
    NumSceneEvents			: 0,
	MaxScroll               : 2,
    LastFixedSceneId        : 2,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 3,
	HasLocalControl			: true,
	HasDirectSceneControl	: false,
	NumStatusLEDs			: 0,
	StatusLEDActionNames    : { "-": "Don't change LED Mode"},
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 1, // Actually 5 but effectively 1 since only the central controller accepts central scene.
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "",
	StatusLEDTimeouts 		: "",
	StatusLEDColors         : "",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1", "2", "3"],
								/* 1 - Scenes  */ []
							  ],
	MaxLabelSize			: 1,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: false,
	SendsBasicSet			: false,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : true,
	DevType                 : DEVTYPE_HOMESEERWD100,
	DeviceXml               : "D_HomeSeerWD100.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:HomeSeerWD100",
	OldParamPrefix          : "HS",

	GetDefaultModeString    : HomeSeer_GetDefaultModeString,
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"O", "T", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 2 : 0) + (taps-1)*4 + (state-1) * 1000;
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Top button","Bottom button"]
	],
	GetRowStyle				: function(button) {return 'style="background-color:'+(button==1?'#F4F4F4':button==2?"#E8E8E8":"#DDFFDD")+';"';},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return labelIndex == 2 ? 0 : 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};

// HomeSeer HS-WS200+ switch-specific
var HOMESEERWS200 = {
	Id						: "HOMESEERWS200",
	Name                    : "HomeSeer HS-WS200+ switch",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 2,
    NumSceneEvents			: 8,
	MaxScroll               : 2,
    LastFixedSceneId        : 2,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 5,
	HasLocalControl			: true,
	HasDirectSceneControl	: false,
	NumStatusLEDs			: 1,
	StatusLEDActionNames    : {
		"-": "Don't change LED Mode", 
		"N": "Set Local Level LED Mode", 
		"L": "Set Device Level LED Mode", 
		"1": "Set Status LED Mode", 
		"A": "Set All LEDs", 
		"R": "Restore previous mode"
	},
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 5,
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "-NL1R",
	StatusLEDTimeouts 		: "12345S",
	StatusLEDColors         : "KRGBMYCWrgbmycw",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1", "2", "3", "4", "5"],
								/* 1 - Scenes  */ ["A", "B", "C", "D", "E", "F", "G", "H" ]
							  ],
	MaxLabelSize			: 1,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: false,
	SendsBasicSet			: false,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : true,
	DevType                 : DEVTYPE_HOMESEERWS200,
	DeviceXml               : "D_HomeSeerWS200.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:HomeSeerWS200",
	OldParamPrefix          : "HS",

	GetDefaultModeString    : HomeSeer_GetDefaultModeString,
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"O", "T", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 2 : 0) + (taps-1)*4 + (state-1) * 1000;
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Top button","Bottom button","Scene Action"]
	],
	GetRowStyle				: function(button) {return 'style="background-color:'+(button==1?'#F4F4F4':button==2?"#E8E8E8":"#DDFFDD")+';"';},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return labelIndex == 2 ? 0 : 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};

// HomeSeer HS-WD200+ dimmer-Specific
var HOMESEERWD200 = {
	Id						: "HOMESEERWD200",
	Name                    : "HomeSeer HS-WD200+ dimmer",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 2,
    NumSceneEvents			: 8,
	MaxScroll               : 2,
    LastFixedSceneId        : 2,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 5,
	HasLocalControl			: true,
	HasDirectSceneControl	: false,
	NumStatusLEDs			: 7,
	StatusLEDActionNames    : {
		"-": "Don't change LED Mode", 
		"N": "Set Local Level LED Mode", 
		"L": "Set Device Level LED Mode", 
		"7": "Set Status LED 7",
		"6": "Set Status LED 6", 
		"5": "Set Status LED 5", 
		"4": "Set Status LED 4", 
		"3": "Set Status LED 3", 
		"2": "Set Status LED 2",
		"1": "Set Status LED 1 (Bottom)", 
		"A": "Set All LEDs", 
		"R": "Restore previous mode"
	},
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 1, // Actually 5 but effectively 1 since only the central controller accepts central scene.
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "-NL7654321AR",
	StatusLEDTimeouts 		: "12345S",
	StatusLEDColors         : "KRGBMYCWrgbmycw",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1", "2", "3", "4", "5"],
								/* 1 - Scenes  */ ["A", "B", "C", "D", "E", "F", "G", "H" ]
							  ],
	MaxLabelSize			: 1,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: false,
	SendsBasicSet			: false,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : true,
	DevType                 : DEVTYPE_HOMESEERWD200,
	DeviceXml               : "D_HomeSeerWD200.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:HomeSeerWD200",
	OldParamPrefix          : "HS",

	GetDefaultModeString    : HomeSeer_GetDefaultModeString,
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"O", "T", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 2 : 0) + (taps-1)*4 + (state-1) * 1000;
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Top button","Bottom button","Scene Action"]
	],
	GetRowStyle				: function(button) {return 'style="background-color:'+(button==1?'#F4F4F4':button==2?"#E8E8E8":"#DDFFDD")+';"';},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return labelIndex == 2 ? 0 : 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};

// HomeSeer HS-FC200+ fan controller-Specific
var HOMESEERFC200 = {
	Id						: "HOMESEERFC200",
	Name                    : "HomeSeer HS-FC200+ fan controller switch",
	DefaultLcdVersion       : 1,
	HasScreen               : false,
	HasMultipleScreens      : false,
	screenWidth             : 1,  // Dummy if not HasScreen
	rightJustifyScreenWidth : 1,
	HasPresetLanguages      : false,
	HasThremostatControl    : false,
    NumButtons              : 2,
    NumSceneEvents			: 8,
	MaxScroll               : 2,
    LastFixedSceneId        : 2,
	HasOffScenes            : false,
	HasIndicator			: false,
	HasMultiTap				: 5,
	HasLocalControl			: true,
	HasDirectSceneControl	: false,
	NumStatusLEDs			: 4,
	StatusLEDActionNames    : {
		"-": "Don't change LED Mode", 
		"N": "Set Local Level LED Mode", 
		"L": "Set Device Level LED Mode", 
		"4": "Set Status LED 4", 
		"3": "Set Status LED 3", 
		"2": "Set Status LED 2",
		"1": "Set Status LED 1 (Bottom)", 
		"A": "Set All LEDs", 
		"R": "Restore previous mode"
	},
	HasCooperConfiguration  : false,
	MaxDirectAssociations   : 1, // Actually 5 but effectively 1 since only the central controller accepts central scene.
	DefaultScreen           : "P1",
	NumTemperatureScreens   : 0,
	StatusLEDActions		: "-NL4321AR",
	StatusLEDTimeouts 		: "12345S",
	StatusLEDColors         : "KRGBMYCWrgbmycw",
	EventCodes              : [
	    						/* 0 - Buttons */ ["1", "2", "3", "4", "5"],
								/* 1 - Scenes  */ ["A", "B", "C", "D", "E", "F", "G", "H" ]
							  ],
	MaxLabelSize			: 1,
	HasBattery				: false,
	HasButtonModes			: false,
	SendsSceneActivation	: false,
	SendsBasicSet			: false,
	SendsSwitchMultilevel   : false,
	SendsCentralScene       : true,
	CentralSceneSlowRefresh : true,
	DevType                 : DEVTYPE_HOMESEERFC200,
	DeviceXml               : "D_HomeSeerFC200.xml",
	OldServiceId            : "urn:gengen_mcv-org:serviceId:HomeSeerFC200",
	OldParamPrefix          : "HS",

	GetDefaultModeString    : HomeSeer_GetDefaultModeString,
	ScreenTypes 			: [
		{ prefix: "P", name: "Preset",      num: 1 }
	],
	CustomModeList 			: [
		"O", "T", "3", "4", "5", "6", "7", "8", "9"
	],
	CustomFontList          : [ ],
	SceneNum : function(screenType, screenNum, virtualButton, state, taps, hold, timeout) {
		return timeout ? 9000 : virtualButton + (hold ? 2 : 0) + (taps-1)*4 + (state-1) * 1000;
	},
	// Returns true if the screen/version combination is supported at least in English
	ScreenIsCompatible		: function(SCObj, screenType, screenNum, version) {
		return screenType == "P";
	},
	// Return a list of languages supported for the screen type/number/version combination
	// indexed into SceneController_PresetLanguages
	LanguagesSupported      : function(screenType, screenNum, version) {
		return [1];
	},
    PresetScreens           : [
		  /* 0  */ null,
		  /* 1  */ ["Top button","Bottom button","Scene Action"]
	],
	GetRowStyle				: function(button) {return 'style="background-color:'+(button==1?'#F4F4F4':button==2?"#E8E8E8":"#DDFFDD")+';"';},
	DefaultDeviceLevel		: function(mode, labelIndex, associationNum) { return labelIndex == 2 ? 0 : 99; },
	IndicatorHighlighting   : "Turn on L.E.D."
};


var ZWDEVICE_SID        = "urn:micasaverde-com:serviceId:ZWaveDevice1";
var PEER_ID             = "PeerID";
var CURRENT_SCREEN      = "CurrentScreen";
var SHOW_SINGLE_DEVICE  = "ShowSingleDevice";
var VERSION_INFO        = "VersionInfo";
var PRESET_LANGUAGE     = "PresetLanguage";
var NUM_LINES		    = "NumLines";
var CAPABILITIES        = "Capabilities";
var EVLCD_DefaultName   = "Evolve LCD Controller";

var SceneController_Select_Tab = 0;
var SceneController_Scenes_Tab = 1;
var SceneController_Copy_Tab = 2;

var SceneController_UI7;
var SceneController_AltUI;
var SceneController_Placeholder = 0;
var SceneController_Placeholder_EventCode = "1";

// return true for UI7, AltUI, false for UI5 or UI6
function SceneController_IsUI7() {
    if (SceneController_UI7 === undefined) {
        SceneController_UI7 = ( "application" in window)	&&
		                 typeof(application) == "object" &&
						 ( "api" in window) &&
						 typeof(api) == "object" &&
		                 typeof(api.cloneObject) == "function";
	}
	return SceneController_UI7;
}

// return true for AltUI
function SceneController_IsAltUI() {
    if (SceneController_AltUI === undefined) {
        SceneController_AltUI = ("VeraBox" in window);
	}
	return SceneController_AltUI;
}

function SceneController_get_device_state(deviceId, service, variable, dynamic) {
	if (SceneController_IsUI7()) {
		return api.getDeviceState(deviceId, service, variable, {dynamic: dynamic});
	} else {
		return get_device_state(deviceId, service, variable, dynamic);
	}
}

function SceneController_get_device_object(deviceID) {
	return SceneController_IsUI7() ? api.getDeviceObject(deviceID) : jsonp.ud.devices[jsonp.udIndex.devices["_"+deviceID]];
}

// Comarison function on table row ID's. Split components by _. Numerics sorted numerically before alphabetics
function SceneController_CompareRowIds(id1, id2) {
	var a1 = id1.split('_');
	var a2 = id2.split('_');
	var i = 0;
	while(true) {
		if (i >= a1.length) {
			if (i >= a2.length) {
				return 0;
			}
			return -1;
		}
		if (i >= a2.length) {
			return 1;
		}
		var s1 = a1[i];
		var s2 = a2[i];
		var n1 = Number(s1);
		var n2 = Number(s2);
		if (isNaN(n1)) {
			if (!isNaN(n2)) {
				return 1;
			}
			if (s1 != s2) {
				return s1 < s2 ? -1 : 1
			}
		} else {
			if (isNaN(n2)) {
				return -1;
			}
			if (n1 != n2) {
				return n1 - n2;
			}
		}
		++i;
	}
}

function SceneController_set_panel_html(html) {
	var $oldTable = jQuery("#ScreenTable tbody")
	if (SceneController_IsUI7()) {
		api.setCpanelContent(html);
	} else {
		set_panel_html(html);
	}
	var $newTable = jQuery("#ScreenTable tbody");
	$newTable.find("td").wrapInner('<div/>');
	if ($oldTable.length > 0) {
		$oldTable.find(".deleted").remove();
		var $oldRow = $oldTable.children().first();
		var $newRow = $newTable.children().first();
		while(1) {
			if ($oldRow.length == 0)	{
				if ($newRow.length == 0) {
					break;
				}
				$newRow.find('td > div').css("display","none").slideDown(500);
				$newRow=$newRow.next();
			} else if ($newRow.length == 0) {
				$oldRow.addClass("deleted").find("td > div").slideUp(500);
				$newTable.append($oldRow);
				$oldRow=$oldRow.next();
			} else {
				var diff = SceneController_CompareRowIds($oldRow[0].id, $newRow[0].id);
				if (diff < 0) {
					var $temp = $oldRow.next();
					$newRow.before($oldRow);
					$oldRow.addClass("deleted").find("td > div").slideUp(500);
					$oldRow=$temp;
				} else if (diff > 0) {
					$newRow.find('td > div').css("display","none").slideDown(500);
					$newRow=$newRow.next();
				} else {
					$oldRow=$oldRow.next();
					$newRow=$newRow.next();
				}
			}
		}
	}
}

// Below is mostly the guts of setUserDataDeviceStateModifyUserData in 1.7.148
// except that we don't call sendCommandSaveUserData.
function SceneController_setUserDataDeviceStateModifyUserData_NoReload(deviceId, service, variable, value) {
    try {
        var deviceObject = application.getDeviceById(deviceId);
        if (typeof(deviceObject) == 'undefined') {
            return false;
        }

        var statesNo;
        var stateObj;
        if (typeof(deviceObject.states) != 'undefined') {
            statesNo = deviceObject.states.length;
            for (i = 0; i < statesNo; i+=1) {
                stateObj = deviceObject.states[i];
                if (typeof(stateObj) != 'undefined' && stateObj.service == service && stateObj.variable == variable) {
                    stateObj.value = value;
                    return true;
                }
            }
        }

        if (typeof(deviceObject.states) == 'undefined') {
            deviceObject.states = [];
        }

        var newState = deviceObject.states.length;
        var maxID = 0;
        if (newState > 0) {
            for (i = 0; i < newState; i+=1) {
                maxID = (deviceObject.states[i].id > maxID) ? deviceObject.states[i].id : maxID;
            }
        }

        deviceObject.states[newState] = {};
        deviceObject.states[newState].service = service;
        deviceObject.states[newState].variable = variable;
        deviceObject.states[newState].value = value;
        deviceObject.states[newState].id = (maxID + 1);
    } catch (e) {
        Utils.logError("Error in SceneController_setUserDataDeviceStateModifyUserData_NoReload: " + e + " " + e.stack);
    }

    return false;
}

function SceneController_set_device_state(deviceId, service, variable, value, persistent) {
	if (SceneController_IsUI7()) {
		api.setDeviceState(deviceId, service, variable, value, {dynamic:true});
		api.setDeviceState(deviceId, service, variable, value, {dynamic:false});
		if (persistent) {
			api.setDeviceStatePersistent(deviceId, service, variable, value, {});
		}
	} else {
		set_device_state(deviceId, service, variable, value, false);
		if (persistent) {
			SceneController_send_action(1,"urn:micasaverde-com:serviceId:HomeAutomationGateway1","SetVariable",{
				"DeviceNum":deviceId, 
				"Service": service, 
				"Variable":variable, 
				"Value":value})
		}
	}
}

function SceneController_IsZWaveChild(deviceId) {
	if ("is_zwave_child" in window) { // UI5, UI6
		return is_zwave_child(deviceId);
	} else if ("isZWaveChild" in api) { // ui7
		return api.isZWaveChild(deviceId);
	} else { // Alternate UI
        var device = api.getDeviceObject(deviceId);
        if (typeof device == "object" && device.id_parent > 0)	{
			var parent = api.getDeviceObject(device.id_parent);
			return typeof parent == "object" && parent.device_type == "urn:schemas-micasaverde-com:device:ZWaveNetwork:1";
		}
		return false;
	}
}

function SceneController_PassthrougZWaveDevice(SCObj, deviceId, functionName) {
	var id = deviceId;
	if(!SceneController_IsZWaveChild(deviceId)) {
	    var peerId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER,PEER_ID, 0));
	    var ShowSingleDevice = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,SHOW_SINGLE_DEVICE, 0);
		if (ShowSingleDevice != "0" || SCObj.HasLocalControl) {
			id = peerId;
		}
	}
	if (functionName in window && typeof(window[functionName]) == "function") { // UI5, UI6
		window[functionName](id);
	} else { // UI7
		myInterface[functionName](id);
	}
}

// Screens page
function EvolveLCD1_Screens(deviceId) {
	SceneController_Screens(EVOLVELCD1, deviceId)
}
function CooperRFWC5_Screens(deviceId) {
	SceneController_Screens(COOPERRFWC5, deviceId)
}
function NexiaOneTouch_Screens(deviceId) {
	SceneController_Screens(NEXIAONETOUCH, deviceId)
}
function HomeSeerWS100_Screens(deviceId) {
	SceneController_Screens(HOMESEERWS100, deviceId)
}
function HomeSeerWD100_Screens(deviceId) {
	SceneController_Screens(HOMESEERWD100, deviceId)
}
function HomeSeerWS200_Screens(deviceId) {
	SceneController_Screens(HOMESEERWS200, deviceId)
}
function HomeSeerWD200_Screens(deviceId) {
	SceneController_Screens(HOMESEERWD200, deviceId)
}
function HomeSeerFC200_Screens(deviceId) {
	SceneController_Screens(HOMESEERFC200, deviceId)
}

// Simple device page pass-through
function EvolveLCD1_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(EVOLVELCD1, deviceId, "simple_device");
}
function CooperRFWC5_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(COOPERRFWC5, deviceId, "simple_device");
}
function NexiaOneTouch_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(NEXIAONETOUCH, deviceId, "simple_device");
}
function HomeSeerWS100_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS100, deviceId, "simple_device");
}
function HomeSeerWD100_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD100, deviceId, "simple_device");
}
function HomeSeerWS200_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS200, deviceId, "simple_device");
}
function HomeSeerWD200_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD200, deviceId, "simple_device");
}
function HomeSeerFC200_simple_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERFC200, deviceId, "simple_device");
}


// Advanced device page pass-through
function EvolveLCD1_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(EVOLVELCD1, deviceId, "advanced_device");
}
function CooperRFWC5_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(COOPERRFWC5, deviceId, "advanced_device");
}
function NexiaOneTouch_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(NEXIAONETOUCH, deviceId, "advanced_device");
}
function HomeSeerWS100_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS100, deviceId, "advanced_device");
}
function HomeSeerWD100_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD100, deviceId, "advanced_device");
}
function HomeSeerWS200_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS200, deviceId, "advanced_device");
}
function HomeSeerWD200_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD200, deviceId, "advanced_device");
}
function HomeSeerFC200_advanced_device(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERFC200, deviceId, "advanced_device");
}


// Z-Wave Options page pass-through
function EvolveLCD1_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(EVOLVELCD1, deviceId, "device_zwave_options");
}
function CooperRFWC5_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(COOPERRFWC5, deviceId, "device_zwave_options");
}
function NexiaOneTouch_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(NEXIAONETOUCH, deviceId, "device_zwave_options");
}
function HomeSeerWS100_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS100, deviceId, "device_zwave_options");
}
function HomeSeerWD100_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD100, deviceId, "device_zwave_options");
}
function HomeSeerWS200_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWS200, deviceId, "device_zwave_options");
}
function HomeSeerWD200_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERWD200, deviceId, "device_zwave_options");
}
function HomeSeerFC200_device_zwave_options(deviceId) {
	SceneController_PassthrougZWaveDevice(HOMESEERFC200, deviceId, "device_zwave_options");
}

// Send an action to the Lua engine
function SceneController_send_action(deviceID,serviceID,action,parameters) {
	if (SceneController_IsAltUI()) {
		api.performActionOnDevice(deviceID, serviceID, action, {actionArguments: parameters});
	} else if (SceneController_IsUI7()) {
		// why doesn't api.performActionOnDevice automatically escape the parameters?
		var encParams = {};
                var kp = Object.keys(parameters);
                var i;
                var kpl = kp.length;
		for (i = 0; i < kpl; i+=1) {
			encParams[kp[i]] = encodeURIComponent(parameters[kp[i]]);
		}
		api.performActionOnDevice(deviceID, serviceID, action, {actionArguments: encParams});
	} else {
		var baseUrl = data_command_url +
	                 (data_command_url.indexOf('/data_request?') > 0 ? '' : '/data_request?');
	               // ui6 data_command_url ends in '/data_request?', ui5 does not.
	    var cmdUrl = baseUrl +
	                 'id=lu_action' +
	                 '&DeviceNum='+deviceID +
	                 '&serviceId='+serviceID +
	                 '&action='+action;
                var param;
		for (param in parameters) {
			cmdUrl += '&' + param + '=' + encodeURIComponent(parameters[param]);
		}
		//console.log("cmdUrl="+cmdUrl);
	    new Ajax.Request(cmdUrl,
	    {
	        method:'get',
	        onSuccess: function(transport){
	            if(transport.responseText.trim() =='' || transport.responseText.indexOf('Error 404')>=0 || transport.responseText.indexOf('ERROR:')>=0) {
					console.log("Scene Controller AJAX error 2: '"+transport.responseText+"' url="+cmdUrl);
	                //show_message("Command failed: '"+transport.responseText+"'",ERROR);
	            }
	        },
	        onFailure: function(transport){
	            if(transport.responseText.indexOf('ERROR:')>=0){
					console.log("Scene Controller AJAX error 2: '"+transport.responseText+"' url="+cmdUrl);
	                //show_message("Command failed: '"+transport.responseText+"'",ERROR);
	            }else{
					console.log("Scene Contoller AJAX error 3: '"+transport.responseText+"' url="+cmdUrl);
	                //show_message("Delivery failed: '"+transport.responseText+"'",ERROR);
	            }
	        }
	    });
	}
}

SceneController_Modes = {
	M:"Momentary",
	D:"Momentary direct",
	T:"Toggle",
	"2":"Two-state",
	"3":"Three-state",
	"4":"Four-state",
	"5":"Five-state",
	"6":"Six-state",
	"7":"Seven-state",
	"8":"Eight-state",
	"9":"Nine-state",
	P:"Mode",
	E:"Energy Mode",
	S:"Toggle Direct",  // Obsolete
	X:"Exclusive",
	N:"Switch Screen",
	H:"Thermostat",
	W:"Welcome",
	C:"Central Scene",
	O:"On",
	L:"Scroll"	// Used internally by Lua. Should never be visible at the JS level.
};

// These are the equivalent mode names for the off button
SceneController_OffModes = {
	"O":"Off",
	"T":"Same as top button",
	"3":"3-state reversed",
	"4":"4-state reversed",
	"5":"5-state reversed",
	"6":"6-state reversed",
	"7":"7-state reversed",
	"8":"8-state reversed",
	"9":"9-state reversed"
}


SceneController_CustomAlignList = [
	"Left",
	"Center",
	"Right"
];

SceneController_PresetLanguages = [
  /* 0 */ null,
  /* 1 */ "English",
  /* 2 */ "Spanish",
  /* 3 */ "Chinese",
  /* 4 */ "German",
  /* 5 */ "French",
  /* 6 */ "Italian",
  /* 7 */ "Punjabi"
];

SceneController_LanguageMaxPresetScreens = [
  /* 0 */ null,
  /* 1 - English */ 26,
  /* 2 - Spanish */  8,
  /* 3 - Chinese */ 10,
  /* 4 - German  */  8,
  /* 5 - French  */  8,
  /* 6 - Italian */  8,
  /* 7 - Punjabi */  8
];

SceneController_TemperatureScreens = [
  /* Temperature 1 - Page 8  */ ["All On", "\u25B2", "72\u00B0", "\u25BC", "All Off"],
  /* Temperature 2 - Page 16 */ ["Lights", "\u25B2", "72\u00B0", "\u25BC", "Privacy"],
  /* Temperature 3 - Page 40 */ ["All On/Off", "\u25B2", "72\u00B0", "\u25BC", "Reading"]
];

SceneController_EventNames = [
 	/* 0 - Buttons */ ["Single Tap", "Double Tap", "Three Taps", "Four Taps", "Five Taps"],
 	/* 1 - Scenes  */ ["A", "B", "C", "D", "E", "F", "G", "H"]
];

SceneController_StatusLEDModeNames = {
	"1": "Show local (single-tap) level", 
	"2": "Show double-tap level", 
	"3": "Show 3-tap level", 
	"4": "Show 4-tap level", 
	"5": "Show 5-tap level", 
	"S": "Show Status mode"
}

SceneController_StatusLEDColorNames = {
	"K": "Off", 
	"R": "Red", 
	"G": "Green", 
	"B": "Blue", 
	"M": "Magenta", 
	"Y": "Yellow", 
	"C": "Cyan", 
	"W": "White",
	"r": "Flashing Red", 
	"g": "Flashing Green", 
	"b": "Flashing Blue", 
	"m": "Flashing Magenta",
	"y": "Flashing Yellow", 
	"c": "Flashing Cyan", 
	"w": "Flashing White" 
}

function SceneController_GetPresetLanguage(SCObj, peerId) {
	if (!SCObj.HasPresetLanguages) {
		return 1;
	}
	var presetLanguage = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, PRESET_LANGUAGE, 0);
	if (!presetLanguage) {
		presetLanguage = "1";	// English
	}
	return parseInt(presetLanguage);
}

function SceneController_SetPresetLanguage(SCObj, peerId) {
	if (!SCObj.HasPresetLanguages) {
		return;
	}
	var presetLanguage=document.getElementById("PresetLanguage_"+peerId).value;
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetPresetLanguage",{Language:presetLanguage});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, PRESET_LANGUAGE,presetLanguage, false);
}


// Called after the user selects an item from the screens pop-up menu
function SceneController_SetScreen(SCObj, peerId) {
	if (!SCObj.HasMultipleScreens) {
		return;
	}
	var curScreen=document.getElementById("CurScreen_"+peerId).value;
	if (!curScreen || typeof curScreen != "string") {
		curScreen = SCObj.DefaultScreen;
	}
	var screenType = curScreen.charAt(0);
	var screenNum = curScreen.slice(1);
	if (SCObj.HasPresetLanguages && (screenType == "P" || screenType == "T")) {
		presetLanguage = SceneController_GetPresetLanguage(SCObj, peerId);
		if (presetLanguage > SceneController_LanguageMaxPresetScreens[screenNum]) {
			presetLanguage = 1;
			SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetPresetLanguage",{Language:presetLanguage});
			SceneController_set_device_state(peerId, SID_SCENECONTROLLER, PRESET_LANGUAGE,presetLanguage, false);
		}
	}
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetScreen", {Screen:curScreen, Timeout:"false", ForceClear:"false"});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, CURRENT_SCREEN, curScreen, false);
	SceneController_Screens(SCObj, peerId);
}

// Work around Vera 7.31 firmware % in variable bug
function SceneController_EscapePercent(str) {
	if (typeof str != "string") {
		return str;
	}
	var s1 = str.replace(/#/g, "##");
	return s1.replace(/%/g, "#p");
}

function SceneController_UnescapePercent(str) {
	if (typeof str != "string") {
		return str;
	}
	var s1 = str.replace(/##/g, "#");
	return s1.replace(/#p/g, "%");
}

// Called after the user has edited a single label
function SceneController_ChangeCustomLabel(SCObj, peerId, screen, labelIndex)
{
	if (SCObj.HasScreen) {
		var text  = document.getElementById( "Text_"+peerId+"_"+screen+"_"+labelIndex);
		var font  = document.getElementById( "Font_"+peerId+"_"+screen+"_"+labelIndex);
		var align = document.getElementById("Align_"+peerId+"_"+screen+"_"+labelIndex);
	    text = text   ? text.value  : "";
		font = font   ? font.value  : "Normal";
		align = align ? align.value : "Center";
		console.log("SceneController_ChangeCustomLabel: label="+text+" font="+font+" align="+align);
		text = SceneController_EscapePercent(text)
		SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateCustomLabel",{Screen:screen,Button:labelIndex,Label:text,Font:font,Align:align});
		SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Label_"+screen+"_"+labelIndex, text, false);
		SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Font_"+screen+"_"+labelIndex, font, false);
		SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Align_"+screen+"_"+labelIndex, align, false);
	}
	SceneController_Screens(SCObj, peerId);
}

function SceneController_ChangeMode(SCObj, peerId, screen, labelIndex, eventCode, item)
{
    console.log("SceneController_ChangeMode: peerId="+peerId+" screen="+screen+" labelIndex="+labelIndex+" eventCode="+eventCode+" item="+item);
    var optEvent = (SCObj.HasMultiTap ||  SCObj.NumSceneEvents > 0) ? ("_" + eventCode) : "" 
    var modePrefixObj  = document.getElementById( "Mode_"+peerId+"_"+screen+"_"+labelIndex+"_"+eventCode);
	var modePrefix = modePrefixObj &&  modePrefixObj.value && typeof  modePrefixObj.value == "string" ? modePrefixObj.value   : "-"
	var modeNegative = false
	if (modePrefix.charAt(0) == "~") {
		modePrefix = modePrefix.charAt(1)
		modeNegative = true
	}
	var oldModeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Mode_"+screen+"_"+labelIndex+optEvent, 0);
	oldModeStr = oldModeStr ? oldModeStr : SCObj.GetDefaultModeString(labelIndex, eventCode);
	var mode = SceneController_ParseModeString(SCObj, oldModeStr, labelIndex, eventCode);
	if (modePrefix == "N") {
		var switchObj = document.getElementById("SwitchScreen_"+peerId+"_"+screen+"_"+labelIndex+'_'+eventCode);
		mode.newScreen = (switchObj ? switchObj.value : (screen == SCObj.DefaultScreen ? "C2" : SCObj.DefaultScreen));
		mode.prefix = "M";
		mode.negative = false;
	} else {
		mode.newScreen=null;
		mode.prefix = modePrefix;
		mode.negative = modeNegative;
	}
	var anyOrAll = document.getElementById("AnyOrAll_"+peerId+"_"+screen+"_"+labelIndex+"_"+eventCode);
	if (anyOrAll) {
		mode.allDevices = anyOrAll.value;
	}
	if (SCObj.NumStatusLEDs > 0) {
    	var ledActionObj  = document.getElementById( "SetLEDAction_"+peerId+"_"+screen+"_"+labelIndex+"_"+eventCode);
		var ledAction = ledActionObj &&  ledActionObj.value && typeof  ledActionObj.value == "string" ? ledActionObj.value   : null;
		var defaultLedColor = "-"
		if (ledAction) {
			defaultLedColor = "W"
			if (SCObj.NumButtons == 2 && labelIndex >= 2 && labelIndex < 3) {
				defaultLedColor = "K"	
			}
		}
		if (anyOrAll) {
			// Ensure that the AnyOrAll selection agrees with whether this is a binary or multilevel indicator/status
			if (ledAction == "L") {
				if (mode.allDevices != "Max" && mode.allDevices != "Min" && mode.allDevices != "Average" && mode.allDevices != "First") {
					mode.allDevices = "Average";
				}
			} else {
				if (mode.allDevices != "Any" && mode.allDevices != "All" && mode.allDevices != "Most") {
					mode.allDevices = "Most";
				}
			}
		}

    	var ledColorObj  = document.getElementById( "LEDColor_"+peerId+"_"+screen+"_"+labelIndex+"_"+eventCode);
		var ledColor = ledColorObj &&  ledColorObj.value && typeof  ledColorObj.value == "string" ? ledColorObj.value   : defaultLedColor;

		var flashPeriodObj = document.getElementById( "FlashPeriod_"+peerId+"_"+screen+"_"+labelIndex+"_"+eventCode);
		var flashPeriod
		if (flashPeriodObj && flashPeriodObj.value) {
			if (flashPeriodObj.value == "") {
				flashPeriod = 5;
			} else {
				var num = Number(flashPeriodObj.value);
				if (num == NaN || num < 0.1 || num > 25.5) {
					window.alert("flash period must be between 0.1 and 25.5 seconds");
					flashPeriod = 5;
				}
				else {
					flashPeriod = Math.floor(num * 10 + .5);
				}
			}
		}

		mode.LEDAction = ledAction;
		mode.LEDColor = ledColor;
		mode.flashPeriod = flashPeriod;
	} else {
		mode.LEDAction = null;
		mode.LEDColor = null;
		mode.flashPeriod = null;
	} 
	var newModeStr=SceneController_GenerateModeString(SCObj, mode, labelIndex, eventCode);
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:screen,Button:labelIndex,EventCode:eventCode,Mode:newModeStr});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER,  "Mode_"+screen+"_"+labelIndex+optEvent, newModeStr, false);
	SceneController_SetQuickTap(SCObj, peerId, screen, labelIndex, eventCode, mode);
	if (item == 1 && mode.prefix >= "2" && mode.prefix <= "9") {
		// Auto fill-in multi-state
		var multiLevel = false
		var numStates = +mode.prefix
		var setStatusColors = SCObj.StatusLEDColors.length > 0 && 
			mode.LEDAction && 
			((mode.LEDAction >= "1" && mode.LEDAction <= "9") || 
			 mode.LEDAction == "A");
		var levels = [];
		for (var i = 0; i < mode.length; ++i) {
			levels[i] = null;
			var info = SceneController_GetDeviceProperties(SceneController_get_device_object(mode[i].device))
			if (info && info.multiLevel) {
				multiLevel = true;
				if (mode[i].level) {
					levels[i] = mode[i].level
					if (levels[i] > 99) {
						levels[i] = 99
					}
				}
			}
		}
		if (multiLevel || setStatusColors) {
			for (var i = 1; i <= numStates; ++i) {
				var stateLabelIndex = labelIndex+(i-1)*1000
				if (multiLevel) {
					for (var j = 0; j < mode.length; ++j) {
						var first = true;
						if (levels[j] != null) {
							if (levels[j] == 0) {
								mode[j].level = Math.floor(i * 100 / (numStates - 1) + .3); /* 0 33 66(not 67) 99 */
							} else {
								mode[j].level = Math.floor(((i + numStates - 2) % numStates) * (levels[j] >= 99 ? 100 : levels[j]) / (numStates - 1) + .3); /* 99 0 33 66 */
							}
							if (mode[j].level > 99) {
								mode[j].level = 99;
							}
							if (screen.charAt(0) == "C" && first) {
								var level = mode[j].level;
								if (level >= 99) {
									level = 100;
								}
								var text = level.toString() + '%';
								var font  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Font_"+screen+"_"+labelIndex);
								var align = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Align_"+screen+"_"+labelIndex);
								font = font   ? font  : "Normal";
								align = align ? align : "Center";
								text = SceneController_EscapePercent(text)
								SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateCustomLabel",{Screen:screen,Button:stateLabelIndex,Label:text,Font:font,Align:align});
								SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Label_"+screen+"_"+stateLabelIndex, text, false);
								SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Font_"+screen+"_"+stateLabelIndex, font, false);
								SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Align_"+screen+"_"+stateLabelIndex, align, false);
								first = false;
							}
						}
					}
				}
				if (setStatusColors) {
					var colorSelect = [
						/* 0 */ "",
						/* 1 */ "W",
						/* 2 */ "WK",
						/* 3 */ "WKG",
						/* 4 */ "WKRC",
						/* 5 */ "WKRMY",
						/* 6 */ "WKRGBC",
						/* 7 */ "WKRGMCY",
						/* 8 */ "WKBRGMCY",
						/* 9 */ "WKbBRGMCY"
					];
					if (levels[1] == 0) {
						mode.LEDColor = colorSelect[numStates].charAt(i % numStates); // K first
					} else {
						mode.LEDColor = colorSelect[numStates].charAt(i-1); // W, K, ...
					}
				}
				var temp = mode.prefix
				var temp2 = mode.sceneId
				var temp3 = mode.offSceneId
				if (i > 1) {
					mode.prefix = "M"
					mode.sceneId = null
					mode.offSceneId = null
				}
				var newStateModeStr=SceneController_GenerateModeString(SCObj, mode, stateLabelIndex, eventCode);
				mode.prefix = temp;
				mode.sceneId = temp2;
				mode.offSceneId = temp3;
				if (i > 1 || newStateModeStr != newModeStr) {
					SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:screen,Button:stateLabelIndex,EventCode:eventCode,Mode:newStateModeStr});
					SceneController_set_device_state(peerId, SID_SCENECONTROLLER,  "Mode_"+screen+"_"+stateLabelIndex+optEvent, newStateModeStr, false);
				}
				SceneController_SetQuickTap(SCObj, peerId, screen, stateLabelIndex, eventCode, mode);
			}
		}
	}
	// If the On button was changed, Handle automatically setting the opposite setting for the off button.
	if (SCObj.NumButtons == 2 && labelIndex >= 1 && labelIndex < 2) {
		var offIndex = labelIndex+1
		var oldOffModeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Mode_"+screen+"_"+offIndex+optEvent, 0);
		oldOffModeStr = oldOffModeStr ? oldOffModeStr : SCObj.GetDefaultModeString(offIndex, eventCode);
		var offMode = SceneController_ParseModeString(SCObj, oldOffModeStr, offIndex, eventCode);
		switch(item) {
			case 1: // Mode
				offMode.prefix = mode.prefix;
				offMode.negative = true;
				break;
			case 2: // LED Action
				if (offMode.prefix == mode.prefix) {
					offMode.LEDAction = mode.LEDAction;
					offMode.LEDColor = "K";
				}
				break;
			case 3: // LED Color
				if (offMode.prefix == mode.prefix) {
					offMode.LEDColor = "K";
				}
				break;
			case 4: // Any Or All (LED Status)
			case 5: // Any Or All (Indicator)
				if (offMode.prefix == mode.prefix) {
					offMode.allDevices = mode.allDevices;
				}
				break;
			case 5: // flash period
				if (offMode.prefix == mode.prefix) {
					offMode.flashPeriod = mode.flashPeriod;
				}
		}
		var newOffModeStr=SceneController_GenerateModeString(SCObj, offMode, offIndex, eventCode);
		SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:screen,Button:offIndex,EventCode:eventCode,Mode:newOffModeStr});
		SceneController_set_device_state(peerId, SID_SCENECONTROLLER,  "Mode_"+screen+"_"+offIndex+optEvent, newOffModeStr, false);
		SceneController_SetQuickTap(SCObj, peerId, screen, offIndex, eventCode, mode);
	}
	SceneController_Screens(SCObj, peerId);
}

// Item=1: checkbox
function SceneController_ChangeNightLightMode(SCObj, peerId, item)
{
    console.log("SceneController_ChangeNightLightMode: peerId="+peerId+" item="+item);
    var enable  = document.getElementById( "NightLightMode_"+peerId).checked;
	var nightLightModeString = enable ? '+' : '-'
	SceneController_send_action(peerId,SID_SCENECONTROLLER, "UpdateNightLightMode",{NightLightMode:nightLightModeString});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "NightLightMode", nightLightModeString, false);
	SceneController_Screens(SCObj, peerId);
}

function SceneController_SetPlaceholder(SCObj, peerId, button, eventCode)
{
	SceneController_Placeholder = button;
	SceneController_Placeholder_EventCode = eventCode;
	SceneController_Screens(SCObj, peerId);
}

function SceneController_AddRemoveEvent(SCObj, peerId, screen, button, add)
{
    console.log("SceneController_AddRemoveEvent: peerId="+peerId+" button="+button+" add="+add);
	var numLines = SCObj.NumButtons
	if (SCObj.NumSceneEvents > 0) {
		numLines++;
	}
	var sceneActionLine = (SCObj.NumSceneEvents > 0 && button == numLines) ? 1 : 0;
	var eventList = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "Events_"+screen+"_"+button, 0);
	var eventList2
	if (!eventList) {
		eventList = SCObj.EventCodes[sceneActionLine][0];
	}
	if (add) {
		for (var i = 0; i < SCObj.EventCodes[sceneActionLine].length; ++i) {
			var eventCode = SCObj.EventCodes[sceneActionLine][i];
			if (eventList.indexOf(eventCode) < 0) {
				eventList += eventCode;
				SceneController_send_action(peerId,SID_SCENECONTROLLER, "SetNumEvents",{Screen:screen,Button:button,Events:eventList});
				SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Events_"+screen+"_"+button, eventList, true);
				if (SCObj.NumButtons == 2 && button == 1) {
					var eventList = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "Events_"+screen+"_"+2, 0);
					if (!eventList) {
						eventList = SCObj.EventCodes[sceneActionLine][0];
					}
					if (eventList.indexOf(eventCode) < 0) {
						eventList += eventCode;
						SceneController_send_action(peerId,SID_SCENECONTROLLER, "SetNumEvents",{Screen:screen,Button:2,Events:eventList});
						SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Events_"+screen+"_"+2, eventList, true);
					}
				}
				if (SCObj.StatusLEDColors.length > 0) {
					var optEvent = "_"+eventCode;
					var oldModeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Mode_"+screen+"_"+button+optEvent, 0);
					if (!oldModeStr) {
						oldModeStr = SCObj.GetDefaultModeString(button, eventCode);
						SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:screen,Button:button,EventCode:eventCode,Mode:oldModeStr});
						SceneController_set_device_state(peerId, SID_SCENECONTROLLER,  "Mode_"+screen+"_"+button+optEvent, oldModeStr, false);
					}
				}
				if (!sceneActionLine) {
					SceneController_Placeholder = button;
					SceneController_Placeholder_EventCode = eventCode;
				}
				break;
			}
		}
	} else {
		if (eventList.length > 1) {
			var eventCode = eventList.slice(eventList.length-1);
			eventList = eventList.slice(0, eventList.length-1)	
			SceneController_send_action(peerId,SID_SCENECONTROLLER, "SetNumEvents",{Screen:screen,Button:button,Events:eventList});
			SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Events_"+screen+"_"+button, eventList, true);
			if (SCObj.NumButtons == 2 && button == 1) {
				eventList2 = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "Events_"+screen+"_"+2, 0);
				if (!eventList2) {
					eventList2 = SCObj.EventCodes[sceneActionLine][0];
				}
				var index = eventList2.indexOf(eventCode);
				if (eventList2.length > 1 && index >= 0) {
					eventList2 = eventList2.slice(0,index) + eventList2.slice(index+1)
					SceneController_send_action(peerId,SID_SCENECONTROLLER, "SetNumEvents",{Screen:screen,Button:2,Events:eventList2});
					SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "Events_"+screen+"_"+2, eventList2, true);
				}
			}

			// Disable the timeout if we are removing the event which we would timeout to.
			var timeoutEnable = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEnable_"+screen, 0) == "true";
			var timeoutEvent = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEvent_"+screen, 0);
			var timeoutScreen = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutScreen_"+screen, 0);
			var timeoutSeconds = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutSeconds_"+screen, 0);
			if (timeoutEnable && (eventList.indexOf(timeoutEvent) < 0 || (SCObj.NumButtons == 2 && button == 1 && eventList2.indexOf(timeoutEvent) < 0))) {
				timeoutEnable = false;
				timeoutEvent = "1"
				SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetScreenTimeout",{ Screen:screen,
																							Enable:String(timeoutEnable),
																							TimeoutScreen:timeoutScreen,
																							TimeoutSeconds:timeoutSeconds,
																							TimeoutEvent:timeoutEvent});
				SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEnable_"+screen, String(timeoutEnable), false);
				SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEvent_"+screen, timeoutEvent, false);
			}
		}
	}
	SceneController_Screens(SCObj, peerId);
}

function SceneController_GetDevice(id)
{
	var i
	for (i = 0; i < jsonp.ud.devices.length; ++i) {
		if (jsonp.ud.devices[i].id == id) {
			return jsonp.ud.devices[i];
		}
	}
	return null
}

// Mode strings consist of ~? Prefix {newScreen:+}? (L..)?(B###)?[AaMmVF]?Q?{(S|C{SceneId@}?{offSceneId@}?)}? {entry}+
// Prefix is M for momentary, T for Toggle, etc and may be preceded by ~ for an "off" negative.
// newScreen is a letter/digit(s) such as C3 for Custom 3 or P12 for Preset 12
// The optional L flag sets the status LED mode and is followed by two characters.
//   The first character is the action: L=normal/Level 1-9=Set status LED, A=Set All LEDs, R=Restore previous mode
//   The second character is the LED color RGBCMYWK (Red ... Black) rgbcmyw (Falshing red ... flashing white), - = No change
// The A (All) or a (Any) flag is present if all/any device matching "on" state should turn LED on.
//   If the A/a flag is not present then if most devices can match the "on" state to turn the LED on.
// The B flag is followed by 3 dicimal digits specifying the blink period in 100ms increments
// For devices with multiple status LEDs, the AnyOrAll flag can take the vales:
//   M (Maximum), m (minimum), V (aVerage) of F (First) to decide the LED level when more than one deives is connected.
// The Q (Quicktap) flag is set if the hold/release events can be ignored and tap events can be acted on immediately.
// The S or C flags indicates that all subsequent associated devices are scene capable
//   and is optinally followed by a sceneId @ which is optional followed by an offSceneId @
// The C flag indicates to use Cooper configuration for non-secene capable devices
// Scene-capable modes are a ; separated list of 0 or more deviceNum,level,dimmingDuration triplets
// Cooper condiguration modes are a ; separate list of 0 or more deviceNum,level pairs
// Non-scene capabile modes are a ; separated list of 0 or more deviceNums
// DeviceNums can be negative if they should only be monitored but not controlled, to set the indicator/status
function SceneController_ParseModeString(SCObj, str, button, eventCode) {
	var mode = [];
	var reResult;
	if (!str) {
		mode.prefix = SCObj.GetDefaultModeString(button, eventCode);
	} else {
		var prefix = str.charAt(0);
		if (prefix == "~") {
			prefix = str.charAt(1)
			mode.prefix = prefix;
			str = str.slice(2);
			mode.negative = true;
		} else {
			mode.prefix = prefix;
			mode.negative = false;
			str = str.slice(1);
		}
		mode.states = 1;
		mode.quickTap = false;
		if (prefix.match(/[2-9]/)) {
			mode.states = parseInt(prefix);
		}
		if (mode.prefix == "S") {
			// Legacy direct toggle. Change into T
			mode.prefix = "T"
		}
		var matchArray = str.match(/([A-Z][0-9]+):+(.*)$/)  // :+ to deal with an old bug
		if (matchArray) {
			mode.newScreen = matchArray[1];
			str = matchArray[2];
		} else if (mode.prefix == "N") {
			// Legacy switch screen.
			mode.prefix = "M"
			mode.newScreen = str;
			return mode;
		}
		if (str.length >= 3 && str.charAt(0) == "L") {
			mode.LEDAction = str.charAt(1)
			mode.LEDColor = str.charAt(2)
			str=str.slice(3);
		}
		if (str.length >= 4 && str.charAt(0) == "B") {
			mode.flashPeriod = parseInt(str.substring(1,4))
			str=str.slice(4); 
		}
		mode.allDevices = null;
		mostMap = {
			"A": "All",
			"a": "Any",
			"M": "Max",
			"m": "Min",
			"V": "Average",
			"F": "First"
		}
		mode.allDevices = mostMap[str.charAt(0)];
		if (mode.allDevices) {
			str = str.slice(1);				
		} else {
			mode.allDevices = "Most";			
		}
		if (str.charAt(0) == "Q") {
			mode.quickTap = true;
			str = str.slice(1);				
		}
		if (str.charAt(0) == "S" || str.charAt(0) == "C") {
			mode.sceneControllable = true;
			if ((reResult = /^.(\d+)@(\d+)@(.*)$/.exec(str)) != null) {
				mode.sceneId = parseInt(reResult[1]);
		 		mode.offSceneId = parseInt(reResult[2]);
				str = reResult[3];
			} else if ((reResult = /^.(\d+)@(.*)$/.exec(str)) != null) {
				mode.sceneId = parseInt(reResult[1]);
				str = reResult[2];
			}
		}
		if (!/;$/.exec(str)) {
			str += ";"
		}
		var re = /(-?\d+)(?:,(\d+)(?:,(\d+))?)?,?;/g
		while ((reResult = re.exec(str))) {
			var device = parseInt(reResult[1])
			var level = reResult[2] ? parseInt(reResult[2]) : null; 
			var dimmingDuration = reResult[3] ? parseInt(reResult[3]) : null; 
			if (device > 0) {
				if (true /* SceneController_GetDevice(device) */) {
					mode.push({device: device,
				           	   level: level,
				           	   dimmingDuration: dimmingDuration})
				}
			} else {
				device = -device;
				if (true /* SceneController_GetDevice(device) */) {
					if (!mode.veraSceneSet) {
					   mode.veraSceneSet=[];	
					}
					mode.veraSceneSet.push({device: device,
				           	   				level: level,
				           	   				dimmingDuration: dimmingDuration})
				}
			}
		}
	}
	return mode;
}

function SceneController_GenerateModeString(SCObj, mode, button, eventCode) {
	if (!mode) {
		mode = [];
	}
	if (!mode.prefix) {
		mode.prefix = SCObj.GetDefaultModeString(button, eventCode);
	}
	var str = mode.prefix;
	if (mode.negative) {
		str = "~" + str
	}
	if (mode.newScreen) {
		str += mode.newScreen + ":"
	}
	var first = true;

	function generateArray(array, sign) {
		if (array) {
			for (var i = 0; i < array.length; ++i) {
				var element = array[i];
				if (element && element.device) {
					if (!first) {
						str += ";";
					}
					if (element.dimmingDuration || element.dimmingDuration == 0) {
						str += element.device*sign + "," + element.level + "," + element.dimmingDuration
					} else if (element.level || element.level == 0) {
						str += element.device*sign + "," + element.level
					} else {
						str += element.device*sign
					}
					first = false;
				}
			}
		}
	}

	if (mode.LEDAction) {
		str += "L" + mode.LEDAction + (mode.LEDColor ? mode.LEDColor : "W");
	}

	if (mode.flashPeriod) {
		str += "B" + ("000" + mode.flashPeriod.toString()).slice(-3);
	}

	var mapMost = {
		"All": "A",
		"Any": "a",
		"Max": "M",
		"Min": "m",
		"Average": "V",
		"First": "F"
	}
	var allKey = mapMost[mode.allDevices]
	if (allKey) {
		str += allKey
	}
	if (mode.quickTap) {
		str += "Q";
	}
	if (mode.sceneControllable) {
		str += "S";
		if (mode.sceneId) {
			str += mode.sceneId + "@"
			if (mode.offSceneId) {
				str += mode.offSceneId + "@"
			}
		}
	}
	generateArray(mode, 1);
	generateArray(mode.veraSceneSet, -1);
	return str;
}

/* ItemChanged: 1=Device Menu, 2=Level Checkbox, 3=Level Input, 4=DimmingDuration Checkbox, 5=DimmingDuration Input */
function SceneController_SelectDirectDevice(SCObj, prefix, peerId, screen, labelIndex, eventCode, associationNum, itemChanged) {
    var optEvent = (SCObj.HasMultiTap ||  SCObj.NumSceneEvents > 0) ? ("_" + eventCode) : "" 
	var modeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Mode_"+screen+"_"+labelIndex+optEvent, 0); // Legacy
	if (!modeStr || typeof modeStr != "string") {
		modeStr = prefix;
	}

	var mode = SceneController_ParseModeString(SCObj, modeStr, labelIndex, eventCode);
	var id='SmartToggle_'+peerId+'_'+screen+'_'+labelIndex+'_'+eventCode+'_'+associationNum;
	var assocDevice = document.getElementById(id).value;
    console.log("SceneController_SelectDirectDevice: peerId="+peerId+" screen="+screen+" eventCode="+eventCode+" associationNum="+associationNum+" assocDevice="+assocDevice);
	if (!assocDevice) {
		assocDevice = "0";
	}

	// If we are changing the first device in the list, it may switch from a scene-capable to a none-scene capable
	// device (or no device) or vice versa. This affects the treatment of the remaining devices on the list.
	if (associationNum == 0) {
		var oldsceneControllable = mode.sceneControllable;
		var affectStart = 1;
		var masterDevice = Number(assocDevice);
		if (assocDevice == "0" && mode.length > 1) {
			masterDevice = mode[1].device;
			affectStart = 2;
		}
		if (masterDevice) {
			var masterObj = SceneController_get_device_object(masterDevice);
			mode.sceneControllable = SceneController_GetDeviceProperties(masterObj).scene;
		}
		else {
			mode.sceneControllable = true;
		}

		if (oldsceneControllable && !mode.sceneControllable && !SCObj.SendsCentralScene) {
			var numAffected = 0;
			var lastName = ""
			for (var i = affectStart; i < mode.length; ++i) {
				if (mode[i] && ((mode[i].level != undefined && mode[i].level != 255) || (mode[i].dimmingDuration != undefined && mode[i].dimmingDuration != 255))) { +
					 +numAffected;
					var obj = SceneController_get_device_object(mode[i].device);
					lastName = obj.name;
				}
			}
			if (numAffected > 0) {
				var levelAnd = SCObj.HasCooperConfiguration ? "" : " level and"
				if (!confirm("You are changing the first device in a direct association list to a non-scene capable device. Are you sure you want to lose all" + levelAnd + " dimming duration settings," +
			                                  " including the " + lastName + "?")) {
					SceneController_Screens(SCObj, peerId);
					return;
				}
			}
		} else if (!oldsceneControllable && mode.sceneControllable && !SCObj.SendsCentralScene) {
			// If we switched from non-scene capable to scene-capable, then remove all non-scene-capable devices.
			var numDelete = 0;
			var lastName = ""
			for (var i = affectStart; i < mode.length; ++i) {
				if (mode[i]) {
					var obj = SceneController_get_device_object(mode[i].device);
					if (SceneController_GetDeviceProperties(obj).basicSetOnly) {
						numDelete++;
						lastName = obj.name;
					}
				}
			}

			if (numDelete > 0 && !confirm("You are changing the first device in a direct association list to a scene capable device. Are you sure you want to remove all non-scene capable devices from the list," +
			                                " including the "+lastName+"?")){
				SceneController_Screens(SCObj, peerId);
				return;
			}

			for (var i = 0; i < mode.length; ++i) {
				if (mode[i]) {
					var obj = SceneController_get_device_object(mode[i].device);
					if (SceneController_GetDeviceProperties(obj).basicSetOnly) {
						mode[i] = null;
					}
				}
			}
		}
	}

	var level = 255
	if (itemChanged == 2) { // Level Checkbox toggled.
		var levelCheckbox = document.getElementById('LevelSelect_'+peerId+'_'+screen+'_'+labelIndex+'_'+eventCode+'_'+associationNum);
		if (levelCheckbox) {
			var levelSelected = levelCheckbox.checked;
			if 	(levelSelected) {
				level = SCObj.DefaultDeviceLevel(mode, labelIndex, associationNum);
			}
		}
	}
	else {
		var levelObject = document.getElementById('Level_'+peerId+'_'+screen+'_'+labelIndex+'_'+eventCode+'_'+associationNum);
		if (levelObject) {
			var levelValue = levelObject.value;
			if (levelValue == "") {
				level = 255;
			} else {
				var num = Number(levelValue);
				if (num == NaN || num < 0 || num > 99 || num != Math.floor(num)) {
					window.alert("Level must be a number between 0 and 99");
					if (mode[associationNum] && mode[associationNum].level) {
						level = mode[associationNum].level;
					}
				}
				else {
					level = num;
				}
			}
		}
	}

	var dimmingDuration = 255
	if (itemChanged == 4) { // Dimming duration Checkbox toggled.
		var dimmingCheckbox = document.getElementById('DimmingDurationSelect_'+peerId+'_'+screen+'_'+labelIndex+'_'+eventCode+'_'+associationNum);
		if (dimmingCheckbox) {
			var dimmingSelected = dimmingCheckbox.checked;
			if 	(dimmingSelected) {
				dimmingDuration = 0;
			}
		}
	}
	else {
		var dimmingObject = document.getElementById('DimmingDuration_'+peerId+'_'+screen+'_'+labelIndex+'_'+eventCode+'_'+associationNum);
		if (dimmingObject) {
			var dimmingValue = dimmingObject.value;
			if (dimmingValue == "") {
				dimmingDuration = 255;
			} else {
				var num = Number(dimmingValue);
				if (num == NaN || num < 0 || num > 254 || num != Math.floor(num)) {
					window.alert("Dimming duration must be a number between 0 and 254");
					if (mode[associationNum] && mode[associationNum].dimmingDuration) {
						dimmingDuration = mode[associationNum].dimmingDuration;
					}
				}
				else {
					dimmingDuration = num;
				}
			}
		}
	}

	assocDevice = parseInt(assocDevice)
	var props = SceneController_GetDeviceProperties(SceneController_get_device_object(assocDevice))
	var sceneControllable = mode.sceneControllable && props && props.scene
	if (!props.multiLevel || (!sceneControllable && !SCObj.HasCooperConfiguration && !SCObj.SendsCentralScene)) {
		level = null;
		dimmingDuration = null;
	} else if ((!props.multiLevelDuration || !SCObj.SendsCentralScene) && !sceneControllable) {
		dimmingDuration = null;
	}
	
	mode[parseInt(associationNum)] = {device: assocDevice, level:level, dimmingDuration:dimmingDuration};
	modeStr = SceneController_GenerateModeString(SCObj, mode, labelIndex, eventCode);

	SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:screen,Button:labelIndex,EventCode:eventCode,Mode:modeStr});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER,  "Mode_"+screen+"_"+labelIndex+optEvent, modeStr, false);
	SceneController_SetQuickTap(SCObj, peerId, screen, labelIndex, eventCode, mode);

	SceneController_Screens(SCObj, peerId);
}

function SceneController_SelectTemperatureDevice(SCObj, peerId, screen)
{
	var temperatureDevice = document.getElementById('TemperatureDevice_'+peerId+'_'+screen).value;
    console.log("SceneController_SelectTemperatureDevice: peerId="+peerId+" screen="+screen+" temperatureDevice="+temperatureDevice);
	if (!temperatureDevice) {
		temperatureDevice = "0";
	}
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateTemperatureDevice",{Screen:screen,TemperatureDevice:temperatureDevice});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TemperatureDevice_"+screen, temperatureDevice, false);
}

function SceneController_PerformAction(SCObj, peerId, action)
{
    console.log("SceneController_PerformAction: peerId="+peerId+" Action="+action);
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"PerformAction",{Action:action});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "LastSceneAction", action, false);
}

function SceneController_ScreenMenu(SCObj, selected, disabled, why, lcdVersion) {
	var html = "";
	for (var i = 0; i < SCObj.ScreenTypes.length; ++i) {
		var screen_type = SCObj.ScreenTypes[i];
		html += '    <optgroup label="' + screen_type.name + '" screens>'
		for (var j = 1; j <= screen_type.num; ++j) {
			if (SCObj.ScreenIsCompatible(SCObj, screen_type.prefix, j, lcdVersion)) {
				var short_name = screen_type.prefix + j;
				html += '     <option value="' + short_name + '"' + (short_name == disabled ? ' disabled' + (why ? ' title="' + why + '" ' : '') : '') +
												 					(short_name == selected ? ' selected' : '') + '>' + screen_type.name + ' ' + j + '</option>';
			}
		}
		html += '    </optgroup>';
	}
	return html;
}

function SceneController_ScreenName(SCObj, screen) {
	if (!screen || screen.length < 2) {
		screen = SCObj.DefaultScreen;
	}
	var screenPrefix = screen.charAt(0);
	var screenNum = screen.slice(1);
	for (var i = 0; i < SCObj.ScreenTypes.length; ++i) {
		var screen_type = SCObj.ScreenTypes[i];
		if (screen_type.prefix == screenPrefix) {
			return screen_type.name + ' ' + screenNum;
		} 
	}
	return "Unknown screen"
}

function SceneController_SortByName(a, b) {
	var x = a.name.toLowerCase();
	var y = b.name.toLowerCase();
	return ((x < y) ? -1 : ((x > y) ? 1 : 0));
}

// Return a sorted menu grouped by room containing all devices that pass the given filter.
// Filter returns:
// 0 - Device not in menu
// 1 - Device is not scene capable. Flagged with a *
// 2 - Device is scene capable. Not flagged
// 3 - Scene capability does not apply.
function SceneController_DeviceMenu(objectID, selectedValue, extraCode, zeroLabel, selectClass, filter){
    var html='';
    var orderedDevices  = JSON.parse(JSON.stringify(jsonp.ud.devices));
    orderedDevices.sort(SceneController_SortByName);
    var orderedRooms  = JSON.parse(JSON.stringify(jsonp.ud.rooms));
    orderedRooms.sort(SceneController_SortByName);
    var somethingSelected = false
    var devicesNO = orderedDevices.length;
    if(devicesNO>0){
        html='<select id="'+objectID+'" '+selectClass+' '+extraCode+'>'

        if(zeroLabel){
        	var selected = selectedValue==0
    		if (selected) {
    			somethingSelected = true 
    		}    	
            html+='<option value="0"' + (selected ? " selected" : "") + '>- '+zeroLabel+' -</option>';
        }
        var roomDevices = {};
		var filterResult;
        for(var i=0;i<devicesNO;i++){
            if(orderedDevices[i].invisible!=1 && (filterResult = filter(orderedDevices[i]))) {
                if(typeof(roomDevices[orderedDevices[i].room])=='undefined'){
                    roomDevices[orderedDevices[i].room] = '';
                }
                var selected = orderedDevices[i].id==selectedValue
	    		if (selected) {
	    			somethingSelected = true 
	    		}    	
                roomDevices[orderedDevices[i].room] += '<option value="'+orderedDevices[i].id+'" '+(selected ?'selected':'')+
													   (filterResult < 3 ?' title="'+(filterResult>1?'This device is scene-capable':'This device is not scene-capable')+'"':'')+
                                                       '>'+orderedDevices[i].name+(filterResult===1?" *":"")+'</option>';
            }
        }

        if (selectedValue > 0 && !somethingSelected) {
               html += '<option value="'+selectedValue+'" selected disabled title="This device was most likely deleted or unpaired."><strong>'+
													   'Missing device #'+selectedValue+'</option>';
        }

        var roomsNO = orderedRooms.length;
        if(typeof(roomDevices[0])!='undefined'){
        	html += '<optgroup label="'+"no room"+'">';
            html += roomDevices[0];
        	html += '</optgroup>';
        }

        for(var i=0;i<roomsNO;i++){
            if(typeof(roomDevices[orderedRooms[i].id])!='undefined'){
            	html += '<optgroup label="'+orderedRooms[i].name+'">';
                html += roomDevices[orderedRooms[i].id];
            	html += '</optgroup>';
            }
        }
 
        html+='</select>';
    }

    return html;
}

function SceneController_DecodeLabel(label, font, align) {
	if (!font) {
		font = "Normal";
	}
	if (!align) {
		align = "Center";
	}
	return {text: label, font: font, align: align};
}

function SceneController_IsDirectZWaveObject(obj) {
	if (!obj || !obj.id_parent) {
		return false;
	}
	return obj.id_parent == 1
}

// Above function returns false for children of Z-Wave multi-channel objects
// Below function returns true in this case.
function SceneController_IsZWaveObject(obj) {
	if (!obj || !obj.id_parent) {
		return false;
	}
	if (obj.id_parent == 1) {
		return true;
	}
	obj = SceneController_get_device_object(obj.id_parent);
	return  obj && obj.id_parent == 1;
}

// This will return:
// zWave: Boolean - Device is Z-Wave
// directZWave: Boolean - Device is a Z-Wave node and not a child of a multi-device Z-Wave node.
// scene: Boolean - Device is Scene Controllable
// basicSetOnly: Boolean - Device is not scene controllable but Basic Set contrllable
// multiLevel: Boolean - Device is multiLevel
// multiLevelDuration: Device is multilevel and also supports a duration attribute using a MULTILEVEL_SET command
// binary: Boolean - Device is binary
function SceneController_GetDeviceProperties(obj) {
	var result = {
		zWave: false,
		directZWave: false,
		scene: false,
		basicSetOnly: false,
		multiLevel: false,
		multiLevelDuration: false,
		binary: false
	}
	if (SceneController_IsZWaveObject(obj)) {
		result.zWave = true;
		if (SceneController_IsDirectZWaveObject(obj)) {
			result.directZWave = true
		}
		var capabilities = SceneController_get_device_state(obj.id, ZWDEVICE_SID, CAPABILITIES, 0);
		if (capabilities) {
			var splitresult = /^[^\|]+\|([\d:,]+)$/.exec(capabilities)
			if (!splitresult || !splitresult[1]) {
				return result;
			}
			var classesString = splitresult[1];
			var re = /(\d+):?(\d*),/g;
			var supportedClasses = {};
			var reResult
			while ((reResult = re.exec(classesString)) != null) {
			   supportedClasses[Number(reResult[1])] = reResult[2] ? Number(reResult[2]) : 1;
			}
			if (supportedClasses[43] && // COMMAND_CLASS_SCENE_ACTIVATION
				supportedClasses[44]) { // COMMAND_CLASS_SCENE_ACTUATOR_CONF) {
				result.scene = true;
			} else {
				result.basicSetOnly = true; // Every Z-Wave device should implicitly support Basic Get/Set.
			}
			if (supportedClasses[38]) { // COMMAND_CLASS_SWITCH_MULTILEVEL
				result.multiLevel = true
				if (supportedClasses[38] >= 2) {
					result.multiLevelDuration = true;
				}
			} 
			if (supportedClasses[37]) {	// COMMAND_CLASS_SWITCH_BINARY
				result.binary = true
			}
		}
	} else { // Non-Z-Wave object
		if (obj.ControlURLs) {
			for (var key in obj.ControlURLs) {
				var cu = obj.ControlURLs[key];
				if (cu.service == "urn:upnp-org:serviceId:SwitchPower1") {
					result.binary = true;
				}
				if (cu.service == "urn:upnp-org:serviceId:Dimming1") {
					result.multiLevel = true;
				}
			}
		}
	}
	return result
}

// Set the QuickTap flag which, if true, we don't need to timeout after a button press to wait for hold/release events
// Since DoubleTap-hold comes as a single tap followed by a hold event, we actually set the flag of the mode for one less tap.
function SceneController_SetQuickTap(SCObj, peerId, screen, stateButton, eventCodePlusOne, mode) {
	if (eventCode < "2" || eventCode > "9") {
		return;
	}
	var eventCode = toString(parseInt(eventCodePlusOne) - 1)
	var optEvent = (SCObj.HasMultiTap ||  SCObj.NumSceneEvents > 0) ? ("_" + eventCode) : "" 
	var modeVar = "Mode_"+screen+"_"+stateButton+optEvent;
	var modeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, modeVar, 0);
	if (!modeStr || typeof modeStr != "string") {
		modeStr = SCObj.GetDefaultModeString(button, eventCode);
	}
	var modeMinusOne = SceneController_ParseModeString(SCObj, modeStr, stateButton, eventCode);
	var oldQuickTap = mode.quickTap

	if (SCObj.HasMultiTap) {
		modeMinusOne.quickTap = true;
		// If any attached devices are multilevel, then no quickTap
		for (var i = 0; i < mode.length; ++i) {
			var info = SceneController_GetDeviceProperties(SceneController_get_device_object(mode[i].device))
			if (info.multiLevel) {
				modeMinusOne.quickTap = false;
				break;
			}
		}
		if (mode.quickTap) {
			// If any Hold or release vera scenes, then no quickTap
			var screenType = screen.charAt(0);
			var screenNum = parseInt(screen.slice(1));
			var state = Math.floor(stateButton/1000) + 1;
			var button = stateButton % 1000; 
			var holdSceneNum = SCObj.SceneNum(screenType, screenNum, button, state, eventCode, true, false)
			if (SceneController_FindScene(peerId, holdSceneNum, 3) /* hold */ ||
				SceneController_FindScene(peerId, holdSceneNum, 4) /* release */) {
				modeMinusOne.quickTap = false;
			}
		}
	} else {
		modeMinusOne.quickTap = false;		
	} 
	if (oldQuickTap != modeMinusOne.quickTap) {
		var newModeStr=SceneController_GenerateModeString(SCObj, modeMinusOne, stateButton, eventCode);
		SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:curScreen,Button:stateButton,EventCode:eventCode,Mode:newModeStr});
		SceneController_set_device_state(peerId, SID_SCENECONTROLLER, modeVar, newModeStr, false);
	}
}

function SceneController_EscapeHTML(string) {
    return string
         .replace(/&/g, "&amp;")
         .replace(/</g, "&lt;")
         .replace(/>/g, "&gt;")
         .replace(/"/g, "&quot;")
         .replace(/'/g, "&#039;");
}

function SceneControllerStyleSheet() {
	return "\
<style> \
.showdisabledtitle:disabled {pointer-events: auto;}	\
</style> ";
}
// Function to draw the Edit screens tab
function SceneController_Screens(SCObj, deviceId) {
	try {
		var html=SceneControllerStyleSheet();
		var peerId
		if(SceneController_IsZWaveChild(deviceId)){
	    	peerId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
		}
		else {
	    	peerId = deviceId;
	    	deviceId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
		}
		var deviceName = SceneController_get_device_object(peerId).name
		var versionInfo = SceneController_get_device_state(deviceId, ZWDEVICE_SID, VERSION_INFO, 0)
		var lcdVersion = versionInfo ? (parseInt(/,(\d+)$/.exec(versionInfo)[1])) : SCObj.DefaultLcdVersion;
		var curScreen = SCObj.HasMultipleScreens ? SceneController_get_device_state(peerId, SID_SCENECONTROLLER, CURRENT_SCREEN, 0) : SCObj.DefaultScreen;
		if (!curScreen || typeof curScreen != "string") {
			curScreen = SCObj.DefaultScreen;
		}
		var screenType = curScreen.charAt(0);
		var screenNum = parseInt(curScreen.slice(1));
		var presetLanguage = SceneController_GetPresetLanguage(SCObj, peerId);
		var hasCustomLabels = (screenType == 'C' || (screenType == 'T' && screenNum > 3));
		var numColumns = hasCustomLabels ? 8 : 6;
		var extraLines = 0;
		var numLines = SCObj.NumButtons

		var timeoutEnable = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEnable_"+curScreen, 0) == "true";
		var timeoutScreen = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutScreen_"+curScreen, 0);
		if (!timeoutScreen) {
			timeoutScreen = SCObj.HasMultipleScreens && curScreen == SCObj.DefaultScreen ? "C2" : SCObj.DefaultScreen;
			timeoutEnable = false;
		}
		var timeoutSeconds = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutSeconds_"+curScreen, 0);
		if (!timeoutSeconds) {
			timeoutSeconds = SCObj.HasMultipleScreens ? 30 : 10;
			timeoutEnable = false;
		}
		var timeoutSceneNum = SCObj.SceneNum(screenType, screenNum, 1, 1, 1, false, true)
		var timeoutEvent = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEvent_"+curScreen, 0);
		if (!timeoutEvent) {
			timeoutEvent = "1"
		}

		var nightLightModeString = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "NightLightMode", 0);
		if (!nightLightModeString) {
			nightLightModeString = "-"
		}
		var NightLightMode = nightLightModeString.charAt(0) == "+";

		var buttonClass = 'class="' + (SceneController_IsUI7() ? 'vBtn showdisabledtitle' : 'btn showdisabledtitle') + '"';
		var selectClass = 'class="' + (SceneController_IsUI7() ? 'select' : 'styled') + '"';
		var inputClass  = 'class="' + (SceneController_IsUI7() ? 'input' : 'styled') + '"';

		if (SCObj.HasMultipleScreens) {
			// Screens pop-up
			html+= '<table style="padding-left:10px;padding-right:10px;" border="0" align="center" class="m_table skinned-form-controls skinned-form-controls-mac">\n' +
			       ' <tr>\n' +
				   '  <th style="text-align:right;"><b>Screen:</b></th>\n' +
		           '  <td>\n' +
	               '   <select '+selectClass+' id="CurScreen_'+peerId+'" onChange="SceneController_SetScreen('+SCObj.Id+','+peerId+')" style="width:120px;">\n' +
		                 SceneController_ScreenMenu(SCObj, curScreen, null, null, lcdVersion) +
		           '   </select>\n' +
		           '  </td>\n';
		    // Number of lines pop-up to enable scrolling
			if (SCObj.MaxScroll > SCObj.NumButtons) {
				if (screenType == "C") {
					var numLinesString = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "NumLines_"+curScreen, 0);
					if (numLinesString) {
						numLines = parseInt(numLinesString);
						numLines = numLines ? numLines : SCObj.NumButtons;
					}
					if (numLines < SCObj.NumButtons) {
						numLines = SCObj.NumButtons;
					}
					if (numLines > SCObj.MaxScroll) {
						numLines = SCObj.MaxScroll;
					}
		            html += '  <th style="text-align:right;width:80px"><b>Lines:</b></th>\n' +
					        '  <td style="width:120px;">\n' +
			                '   <select '+selectClass+' id="NumLines_'+peerId+'_'+curScreen+'" onChange="SceneController_ChangeNumLines('+SCObj.Id+','+peerId+',\''+curScreen+'\')" style="width:100px;">\n';
					for (i = SCObj.NumButtons; i <= SCObj.MaxScroll; ++i) {
						html += '     <option value="' + i + '"' + (i==numLines ? ' selected' : '') + '>' + i + '</option>\n';
					}
				    html += '   </select>\n' +
				            '  </td>\n';
				} else {
					html += '  <th style="width:80px;"></th><td style="width:120px;"></td>\n';
				}
			}
			// Languages pop-up
			var languageList = SCObj.LanguagesSupported(screenType, screenNum, lcdVersion);
			if (languageList.length > 1) {
	            html += '  <th style="text-align:right;width:80px"><b>Language:</b></th>\n' +
				        '  <td style="width:120px;">\n' +
		                '   <select '+selectClass+' id="PresetLanguage_'+peerId+'" onChange="SceneController_SetPresetLanguage('+SCObj.Id+','+peerId+')" style="width:100px;">\n';
				for (i = 1; i <= languageList.length; ++i) {
					html += '     <option value="' + i + '"' + (i==presetLanguage ? ' selected' : '') + '>' + SceneController_PresetLanguages[languageList[i-1]] + '</option>\n';
				}
			    html += '   </select>\n' +
			            '  </td>\n';
			}
			else {
				html += '  <th style="width:80px;"></th><td style="width:120px;"></td>\n';
			}
			html +=  ' </tr>\n' +
			         '</table>\n' +
			         '<div class="m_separator_inner"></div>\n';
		}

		if (SCObj.NumSceneEvents > 0) {
			numLines++;
		}
		// Main table
		html += '<table id="ScreenTable" align="left" style="text-align:center;padding-left:10px;padding-right:10px;white-space:nowrap;border-spacing:0px" border="0" align="center" class="m_table skinned-form-controls skinned-form-controls-mac">\n';
		if (screenType == "C" || screenType == "P" || screenType == "T") {
			// Header row
			html += ' <tr id="Screeen_A">\n';
			html += '  <th style="text-align:left;width:10px;"><b>Line</b></th>'
			if (SCObj.HasScreen) {
				html += '  <th style="text-align:center;width:10px;"><b>Label</b></th>';
			    if (hasCustomLabels) {
			    	html += '  <th style="text-align:center;width:10px;"><b>Font</b></th>\n'
			    	html += '  <th style="text-align:center;width:10px;"><b>Align</b></th>\n'
			    }
			} else {
				html += '  <th style="text-align:center;width:10px;"><b>Button</b></th>\n';				
			}
			if (SCObj.HasMultiTap > 0) {
				html += '  <th style="text-align:center;width:10px;"><b>Event</b></th>\n'
			} 
			html += '  <th style="text-align:center;width:110px;"><b>Mode</b></th>\n'
			html += '  <th style="text-align:center;width:10px;"><b>Vera Scene</b></th>\n'
			html += '  <th style="text-align:center;width:100px;"><b>Add/Remove Line</b></th>\n'
			html += ' </tr>\n';

			// One line for each button
			for (var button = 1; button <= numLines; ++button) {
				var label;
				var font = "";
				var align = "";
				var custom = false;
				var rowStyle = SCObj.GetRowStyle(button)
				var sceneActionLine = (SCObj.NumSceneEvents > 0 && button == numLines) ? 1 : 0;
				var eventList = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "Events_"+curScreen+"_"+button, 0);
				if (!eventList) {
					eventList = SCObj.EventCodes[sceneActionLine][0];
				}
				if (SCObj.NumStatusLEDs > 0 && button > SCObj.NumButtons) {
					/* LED status timeout choice */
					html += ' <tr id="Screen_C_'+button+'_0_0">\n'+
					        '  <td align="left" colspan=' + numColumns + '>\n'+
						    '   <input type="checkbox"'+(timeoutEnable ?' checked':'')+' id="TimeoutEnable_'+peerId+'_'+curScreen+'" style="min-width:'+(SceneController_IsUI7() ? 20 : 10)+'px;"'+
						          ' onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', false)" /><span></span>\n'+
						    '   User button actions revert to \n'+
				            '   <select '+selectClass+' id=TimeoutEvent_'+peerId+'_'+curScreen+
				                  ' onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', true)">\n';
				    for (var k = 0; k < SCObj.StatusLEDTimeouts.length; ++k) {
				    	var timeoutEventItem = SCObj.StatusLEDTimeouts.charAt(k);
						var show = true;
						if (timeoutEventItem >= "1" && timeoutEventItem <= "9" && timeoutEventItem != timeoutEvent) {
							// Enable Revert to N-tap level only if at least one button goes up to N taps or if it is already selected.
							show = false;
							for (var L = 1; L <= SCObj.NumButtons; ++L) {
								var eventList2 = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,  "Events_"+curScreen+"_"+L, 0);
								if (eventList2 && eventList2.search(timeoutEventItem) >= 0) {
									show = true;
									break;
								}
							}
						}
						if (timeoutEventItem == "A") {
							show = false  // Set all LED's is covered by Show status mode
						}
						if (show) {
				    		html += '    <option value="'+timeoutEventItem+'"'+(timeoutEventItem == timeoutEvent ? ' selected' : '')+'>'+
				    		        SceneController_StatusLEDModeNames[timeoutEventItem]+'</option>\n';
				    	}
				    }
					html += '   </select> mode after \n'+
			                '   <input '+inputClass+' id="TimeoutSeconds_'+peerId+'_'+curScreen+'" type="number" value="'+timeoutSeconds+'" style="width:40px;"'+
			                    ' onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', 3)" />\n'+
			        		'   seconds.\n'+
			        		'  </td>\n'+
			        		' </tr>\n';

					/* Night light mode */
					html += ' <tr id="Screen_C_'+button+'_0_1">\n'+
					        '  <td align="left" colspan=' + numColumns + '>\n'+
					        '   <span title="If checked, a LED is on when switch is off.">\n'+
						    '     <input type="checkbox"'+(NightLightMode ?' checked':'')+' id="NightLightMode_'+peerId+'" style="min-width:'+(SceneController_IsUI7() ? 20 : 10)+'px;"'+
						    '       onChange="SceneController_ChangeNightLightMode('+SCObj.Id+','+peerId+', 1)" /><span></span>\n'+
						    '     Night Light mode \n'+
						    '   </span>\n'
			        		'  </td>\n'+
			        		' </tr>\n';

					html += ' <tr id="Screen_C_'+button+'_0_2"><td><div style="height: 10px;"/></td></tr>\n';
				}
				/* Loop through all events (click-types) for a given button */
				for (var eventNum = 1; eventNum <= eventList.length; ++eventNum) {
					var eventCode = eventList.charAt(eventNum-1)
					var optEvent = (SCObj.HasMultiTap ||  SCObj.NumSceneEvents > 0) ? ("_" + eventCode) : "" 
					var modeVar = "Mode_"+curScreen+"_"+button+optEvent;
					var modeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, modeVar, 0);
					if (!modeStr || typeof modeStr != "string") {
						modeStr = SCObj.GetDefaultModeString(button, eventCode);
					}
					var mode = SceneController_ParseModeString(SCObj, modeStr, button, eventCode);
					var modePrefix = mode.prefix;
					var modeNegative = mode.negative;
					var states = mode.states;
					if (mode.negative) {
						states = 1;
					}
					if (screenType == "C") {
						label = SceneController_UnescapePercent(SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Label_"+curScreen+"_"+button, 0));
						font = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Font_"+curScreen+"_"+button, 0);
						align = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Align_"+curScreen+"_"+button, 0);
						custom = true;
					} else if (screenType == "T") {
						if (screenNum <= 3) {
							label = SceneController_TemperatureScreens[screenNum-1][button-1];
						}
						else {
							if (button >= 2 && button <= 4) {
								label = SceneController_TemperatureScreens[0][button-1];
							} else if (modePrefix == "P") {
								label = "Heat/Cool/Auto/Off";
							} else if (modePrefix == "E") {
								label = "Normal/Energy Saving"
							} else {
								label = SceneController_UnescapePercent(SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Label_"+curScreen+"_"+button, 0));
								font = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Font_"+curScreen+"_"+button, 0);
								align = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Align_"+curScreen+"_"+button, 0);
								custom = true;
							}
						}
					} else {
						label = SCObj.PresetScreens[screenNum][button-1];
					}
					if (!label && label !== 0) {
						label = "";
					}

					// Each button has 1 or more events. Each event has one of more states
					for (var state = 1; state <= states; ++state) {
						var stateButton = button + (state-1) * 1000;
						if (state > 1) {
							if (SCObj.HasScreen) {
								label = SceneController_UnescapePercent(SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Label_"+curScreen+"_"+stateButton, 0));
								font = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Font_"+curScreen+"_"+stateButton, 0);
								align = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "Align_"+curScreen+"_"+stateButton, 0);
								if (!label) {
									label = "";
								}
							}
							modeVar = "Mode_"+curScreen+"_"+stateButton+optEvent;
							modeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, modeVar, 0);
							if (!modeStr || typeof modeStr != "string") {
								modeStr = SCObj.GetDefaultModeString(stateButton, eventCode);
							}
						}
						mode = SceneController_ParseModeString(SCObj, modeStr, button, eventCode)
						if (state > 1) {
							mode.prefix = "M"
							mode.negative = false;
						}

						// show line as button{_evnet}{.state}
						html += ' <tr id="Screen_C_'+button+'_'+eventNum+'_'+state+'" '+rowStyle+'>\n';
						if (sceneActionLine) {
							html += '<td>'+eventCode+'</td>'
						} else {
							html += '  <td>'+button;
							if (eventNum > 1) {
								html += '_'+eventCode
							}
							if (states > 1) {
								html +=  '.'+state;
							}
							html += '</td>\n';
						}

						// Display the custom label
						if (custom) {
							var spec = SceneController_DecodeLabel(label, font, align);
							html += '  <td><input style="text-align:'+spec.align + ';' +
							                            ' width:100px;' +
								                        ' font-size: ' + (spec.font=='Large' ? '150%' : '100%') + ';' +
							                            ' letter-spacing:' + (spec.font=='Compressed' ? '-1px' : '1px') + ';' +
							                            ' color:' + (spec.font=='Inverted' ? 'white':'black') + ';' +
							                            ' background-color:' + (spec.font=='Inverted' ? 'black':'white') + ';"' +
							               ' type="text"'+
							               ' id="Text_'+peerId+'_'+curScreen+'_'+stateButton+'" value="'+spec.text/*.replace(/\\/g,'\\\\')*/+'"' +
								           ' title=' + (spec.font == 'Normal' ? '"Use a - or \\r for a two-line label in normal text"' : '"Only one line per lable in ' + spec.font.toLowerCase() + ' text"') +
							               ' onChange="SceneController_ChangeCustomLabel('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+')"></td>\n';
							// Font
							html += '  <td style="text-align:left;column-width:'+(SceneController_IsUI7() ? 120 : 90)+'px;">\n'+
							        '   <select '+selectClass+' style="height:22px;"' +
							           ' id="Font_'+peerId+'_'+curScreen+'_'+stateButton+'"'+
							           ' onChange="SceneController_ChangeCustomLabel('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+')">\n';
							for (var j = 0; j < SCObj.CustomFontList.length; ++j) {
								var fontName = SCObj.CustomFontList[j];
								html += '    <option style="height:22px;" value="'+fontName+'" '+(fontName == spec.font ? 'selected' : '')+'>'+fontName+'</option>\n';
							}
							html += '   </select>\n'+
							        '  </td>\n';
							// Align
							html += '  <td style="text-align:left;column-width:'+(SceneController_IsUI7() ? 90 : 60)+'px;">\n'+
							        '   <select '+selectClass+' style="text-align:left; height:22px;" id="Align_'+peerId+'_'+curScreen+'_'+stateButton+'"' +
							        ' onChange="SceneController_ChangeCustomLabel('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+')">\n';
							for (var j = 0; j < SceneController_CustomAlignList.length; ++j) {
								var alignName = SceneController_CustomAlignList[j];
								html += '    <option style="height:22px;" value="'+alignName+'" '+(alignName == spec.align ? 'selected' : '')+'>'+alignName+'</option>\n';
							}
							html += '   </select>\n'+
							        '  </td>\n';
						}
						else {  // Fixed button name for devces without a screen
							html += '  <td>'+label+'</td>\n';
							if (hasCustomLabels) {
								html += '  <td/>\n  <td/>';
							}
						}
						// Event name or pop-up
						if (SCObj.HasMultiTap) {
							html += '  <td style="text-align:left;column-width:135px;">\n'
							html += '   '+SceneController_EventNames[sceneActionLine][eventNum-1]+'\n';
							html += '  </td>\n';
						}
						var tapSceneNum  = SCObj.SceneNum(screenType, screenNum, button, state, eventNum, false, false)
						var holdSceneNum = SCObj.SceneNum(screenType, screenNum, button, state, eventNum, true, false)
						// Mode pop-up
						if (screenType != "T"  || button == 1 || button == 5) {
							if (sceneActionLine) {
								var disableApplyAction = false;
								var ActionSpecified = false;
								html += '  <td align="center">\n';
								html += '   <button type="button" '+buttonClass+' '+(disableApplyAction?'disabled ':'')+'style="min-width:10px;'+
											             (ActionSpecified?';color:orange;':'')+(disableApplyAction?'background-color:#AAAAAA;':'')+'"'+
														 ' title="Set LED color now."'+
											             ' onClick="SceneController_PerformAction('+SCObj.Id+','+peerId+',\''+eventCode+'\')">Apply</button>\n';
								html += '  </td>\n';
							} else if (state == 1) {
								if (SCObj.HasLocalControl && eventNum == 1) {
									html += '  <td style="text-align:left;">\n';
									if (button == 1) {
										html += 'Local load on'
									} else {
										html += 'Local load off'										
									}	
								} else {
									html += '  <td style="text-align:left;column-width:'+(SceneController_IsUI7() ? 120 : 100)+'px;">\n';
									// Mode-type pop-up
									html += '   <select '+selectClass+
									            ' id="Mode_'+peerId+'_'+curScreen+'_'+button+'_'+eventCode+'"'+
									            ' style="text-align:left;"'+
									            ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+button+',\''+eventCode+'\',1)">\n';
									var start = (SCObj.NumButtons == 2 && button == 2) ? -1 : 0;
									for (var j = start; j < SCObj.CustomModeList.length; ++j) {
										var optionPrefix;
										var optionNegative = false
										if (j < 0) {
											// Off complement to On button.
											var onModeVar = "Mode_"+curScreen+"_1"+optEvent;
											var onModeStr  = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, onModeVar, 0);
											if (!onModeStr || typeof onModeStr != "string") {
												onModeStr = SCObj.GetDefaultModeString(1, eventCode);
											}
											var onMode = SceneController_ParseModeString(SCObj, onModeStr, button, eventCode);
											optionPrefix = onMode.prefix;
											optionNegative = true;
										} else {
											optionPrefix = SCObj.CustomModeList[j];
										}
										var selected = optionPrefix == (mode.newScreen ? "N" : mode.prefix) && optionNegative == mode.negative
										var enabled = true
										if (optionPrefix == "P" || optionPrefix == "E") {
											enabled = screenType == "T" && screenNum > SCObj.NumTemperatureScreens;
										} else if (screenType == "P") {
											//// Preset screens don't have 3+state buttons unless there are enough status LED colors to show all of the states.
											// enabled = optionPrefix < "2" || optionPrefix > "9" || SCObj.StatusLEDColors.length >= parseInt(optionPrefix); 
											enabled = true;  // Allow multi-state even if there are not enough status LED colots to show the current state.
										} 
										if (button > 1 && optionPrefix == "O" && !optionNegative) {
											enabled = false; // Don't enable "On" in the bottom button.
										}
										if (enabled) {
											var optionLabel = (optionNegative && SceneController_OffModes[optionPrefix]) ? SceneController_OffModes[optionPrefix] 
											                                                                             : SceneController_Modes[optionPrefix]; 
											html += '    <option style="height:22px;" value="'+(optionNegative?'~':'')+optionPrefix+'" '+(selected ? 'selected' : '')+'>'+optionLabel+'</option>\n';
										}
									}
									html += '   </select>\n';
								}
								html += '  </td>\n'
							}
							else {
								html += '  <td/>\n';
							}
							// Scene buttons
							var disableMoreDirectScenes = false
							if (sceneActionLine || (mode.negative && mode.prefix != "O")) {
								html += '  <td/>\n'
							} else {	
							    html += '  <td align="left" style="padding-left:3px;"><div style="w1dth:'+ (SceneController_IsUI7() ? 180 : 110) + 'px;'+(SceneController_IsUI7() ? 'padding-top:2px;padding-bottom:2px;': '')+'">\n';
								if (mode.newScreen) {
									var modeParam = mode.newScreen;
									if (!modeParam) {
										modeParam = (curScreen == SCObj.DefaultScreen ? "C2" : SCObj.DefaultScreen);
									}
									html += '   <select '+selectClass+' style="text-align:left; width:120px; height:22px;" id="SwitchScreen_'+peerId+'_'+curScreen+'_'+button+'" onChange="SceneController_ChangeCustomMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+button+')" style="w1dth:100px;">\n' +
									             SceneController_ScreenMenu(SCObj, modeParam, curScreen, "You cannot switch to the screen that you are  already on", lcdVersion) +
									        '   </select>\n';
								}
								var disableVeraScene = false;
								if (mode.length >= SCObj.MaxDirectAssociations && mode.prefix != "C" && SCObj.HasDirectSceneControl) {
									disableVeraScene = SCObj.MaxDirectAssociations + " devices are already connected to this button"
								} else if (mode.prefix == "T" &&  !SCObj.HasOffScenes && !SCObj.HasIndicator && !SCObj.SendsCentralScene) {
									disableVeraScene = "This controller cannot toggle Vera scenes"
								} else if (mode.length > 0 && mode.prefix != "C" && SCObj.HasDirectSceneControl && SceneController_GetDeviceProperties(SceneController_get_device_object(mode[0].device)).basicSetOnly && !SCObj.HasIndicator) {
									disableVeraScene = "The first directly connected device is not scene-capable";
								}
								var veraSceneObjs = SceneController_FindScene(peerId, tapSceneNum, -1)
								if (mode.negative) {
									disableMoreDirectScenes = "The bottom button controls the same devices as the top button"
								} else if (mode.prefix != "C" && SCObj.HasDirectSceneControl) { // Central scene "Direct associations" are not really direct and thus there is no limit.
									if (veraSceneObjs && mode.length >= SCObj.MaxDirectAssociations-1) {
										disableMoreDirectScenes = "Maximum " + (SCObj.MaxDirectAssociations-1) + (SCObj.MaxDirectAssociations==2 ? " device" : " devices") + " already directly connected along with a Vera scene";
									}
									else if (mode.length >= SCObj.MaxDirectAssociations) {	
										disableMoreDirectScenes = "Maximum " + SCObj.MaxDirectAssociations + " devices already directly connected";
									}
								}
								var enableNonSceneDirect = !disableMoreDirectScenes &&  // Cannot enable non-scene direct if direct is disabled.
														   (SCObj.HasCooperConfiguration || 	// Cooper RFWC5 can handle both Scen and non-scene capable devices simultaneously
														    mode.prefix == "C" ||  // Central scene "Direct" devices are not really direct and are handled by the plug-in Lua code
														    ((mode.prefix == "T" || mode.prefix == "X") && (SCObj.HasIndicator || SCObj.SendsCentralScene)) ||  // If we have an indicator, we can determine state in toggle or exclusive mode
														    (!veraSceneObjs && states == 1)); // If we have no need to determine state
								switch (mode.prefix) {
									default:
										if ((SCObj.HasIndicator || SCObj.SendsCentralScene) && ((modePrefix >= "2" && modePrefix <= "9") || mode.prefix == "X"))  {
											// For multistate and exclusive modes, we look at all of the devices affected by each state
											// to allow the indicator to automatically track which state is most appropriate.
											var sceneSet = SceneController_GetSceneSet(veraSceneObjs)
											if (!SceneController_SceneSetsEqual(sceneSet, mode.veraSceneSet)) {
												mode.veraSceneSet = sceneSet;
												var newModeStr=SceneController_GenerateModeString(SCObj, mode, stateButton, eventCode);
												SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:curScreen,Button:stateButton,EventCode:eventCode,Mode:newModeStr});
												SceneController_set_device_state(peerId, SID_SCENECONTROLLER, modeVar, newModeStr, false);
												SceneController_SetQuickTap(SCObj, peerId, curScreen, stateButton, eventCode, mode);
											}
										}
										html += '   <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (SceneController_FindScene(peerId, tapSceneNum, 2)?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"' +
													 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+tapSceneNum+' is activated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+tapSceneNum+',2,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">Scene</button>\n';
								      	break;
								   	case "T":  // Toggle
										var onSceneObjs = SceneController_FindScene(peerId, tapSceneNum, 1)
										if (SCObj.HasIndicator || SCObj.SendsCentralScene) {
											// For toggle modes, we look at all of the devices affected by any Vera "on" scene and used them to
											// to allow the indicator to automatically track when most of the affected devices are on or off.
											var onSceneSet = SceneController_GetSceneSet(onSceneObjs)
											if (!SceneController_SceneSetsEqual(onSceneSet, mode.veraSceneSet)) {
												mode.veraSceneSet = onSceneSet;
												var newModeStr=SceneController_GenerateModeString(SCObj, mode, stateButton, eventCode);
												SceneController_send_action(peerId,SID_SCENECONTROLLER,"UpdateMode",{Screen:curScreen,Button:stateButton,EventCode:eventCode,Mode:newModeStr});
												SceneController_set_device_state(peerId, SID_SCENECONTROLLER, modeVar, newModeStr, false);
												SceneController_SetQuickTap(SCObj, peerId, curScreen, stateButton, eventCode, mode);
											}
										}
										html += '   <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (onSceneObjs?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"'+
													 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+tapSceneNum+' is activated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+tapSceneNum+',1,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">On...</button>\n';
										html += '    <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (SceneController_FindScene(peerId, tapSceneNum, 0)?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"'+
												 	 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+tapSceneNum+' is deactivated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+tapSceneNum+',0,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">Off...</button>\n';
										break;
									case "C":  // Central Scene
									case "O":  // On/Off
										html += '    <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (SceneController_FindScene(peerId, tapSceneNum, 2)?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"'+
												 	 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+tapSceneNum+' is activated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+tapSceneNum+',2,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">Tap...</button>';
										html += '    <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (SceneController_FindScene(peerId, holdSceneNum, 3)?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"'+
												 	 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+holdSceneNum+' is activated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+holdSceneNum+',3,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">Hold...</button>\n';
										html += '    <button type="button" '+buttonClass+' '+(disableVeraScene?'disabled ':'')+'style="min-width:10px;'+
										             (SceneController_FindScene(peerId, holdSceneNum, 4)?';color:orange;':'')+(disableVeraScene?'background-color:#AAAAAA;':'')+'"'+
												 	 ' title="'+(disableVeraScene ? disableVeraScene:'Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+holdSceneNum+' is deactivated')+'"'+
										             ' onClick="SceneController_SetScene('+peerId+','+holdSceneNum+',4,'+eventNum+',\''+SceneController_EscapeHTML(label)+'\')">Release...</button>\n';
										break;
								}
								html += '   </div>\n'
								html += '  </td>\n'
							}
							// Add Direct/Device buttons
							html += '  <td style="text-align:left;">\n'
							if (sceneActionLine || mode.negative) {
								html += '    <div style="width:77px;display:inline-block;"/>\n';
							} else {
								html += '    <button type="button" '+buttonClass+' '+(disableMoreDirectScenes?'disabled title="'+disableMoreDirectScenes+'" ':'')+'style="min-width:10px;'+
									             (mode.length>0?';color:orange;':'')+(disableMoreDirectScenes?'background-color:#AAAAAA;':'')+'"'+
												 (disableMoreDirectScenes?' title="'+disableMoreDirectScenes+'"':'')+
									             ' onClick="SceneController_SetPlaceholder('+SCObj.Id+','+peerId+','+stateButton+',\''+eventCode+'\')">'+
									             ((!SCObj.HasDirectSceneControl || mode.prefix=="C")?'<b>+</b>Device':'+Direct')+'</button>\n';
							}
							// Add/remove event buttons
							if (SCObj.HasMultiTap && eventNum == 1) {
								var disableAddEvent= ((eventList.length < SCObj.EventCodes[sceneActionLine].length))?false:"All events already added";
								var eventActionLabel = sceneActionLine ? "Action" : "Event"
								html += '    <button type="button" '+buttonClass+' '+(disableAddEvent?'disabled title="'+disableAddEvent+'" ':'')+'style="min-width:10px;'+
												 (disableAddEvent?'background-color:#AAAAAA;':'')+'"'+
									             ' onClick="SceneController_AddRemoveEvent('+SCObj.Id+','+peerId+',\''+curScreen+'\','+button+',true)">'+
									             '<b>+</b>'+eventActionLabel+'</button>\n';
								var disableRemoveEvent =  (eventList.length > 1) ?false:"Cannot remove the last event";
								html += '    <button type="button" '+buttonClass+' '+(disableRemoveEvent?'disabled title="'+disableRemoveEvent+'" ':'')+'style="min-width:10px;'+
												 (disableRemoveEvent?'background-color:#AAAAAA;':'')+'"'+
									             ' onClick="SceneController_AddRemoveEvent('+SCObj.Id+','+peerId+',\''+curScreen+'\','+button+',false)">'+
									             '<b>-</b>'+eventActionLabel+'</button>\n';
							}
							html += '  </td>\n';
						    html += ' </tr>\n';
							// Extra lines for direct association
							if (!disableMoreDirectScenes && SceneController_Placeholder == stateButton && SceneController_Placeholder_EventCode == eventCode) {
								mode.push({device:0});
							}
							if (mode.negative && button > 1) {
								mode.push({device:-1});
							}
							for (var j = 0; j < mode.length; ++j) {
								extraLines++;
								html +=  ' <tr id="Screen_C_'+button+'_'+eventNum+'_'+state+'_'+j+'" '+rowStyle+'>\n'+
							             '  <td/><td colspan=' + numColumns + ' align="left">Device: ';
							    if (mode[j].device == -1) {
							    	html += ' Same as top button'
							    } else {
									html +=  SceneController_DeviceMenu('SmartToggle_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'_'+j,
				  	 					mode[j].device,
				  	 					'style="width:200px; height:22px;" onChange="SceneController_SelectDirectDevice('+SCObj.Id+',\''+mode.prefix+'\','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',\''+j+'\',1)"',
				  	 					" ", selectClass,
				  	 					function(obj) {
											for (var k = 0; k < mode.length; ++k) {
												if (k != j && mode[k].device == obj.id) {
													// Already in the list.
													return 0;
												}
											}
											var props = SceneController_GetDeviceProperties(obj);
											if (!props.directZWave && SCObj.HasDirectSceneControl && mode.prefix != "C") { 
												// We allow non-Z-Wave devices amd Multi-device Z-Wave child devices only in Central Scene mode where they are handled specially
												return 0
											}
											if (mode.prefix == "H") {
												if (obj.category_num != 5 && obj.category_num != 17) {
											    // Only allow temperature devices in thermostat mode.
													return 0
												}
											} else {
												// Non-thermostats must be controllable as either binary or multilevel or else basic set.
												if (!props.basicSetOnly && !props.binary && !props.multiLevel) {
													return 0
												}
											}
											if (!SCObj.HasDirectSceneControl || mode.prefix == "C") {
												return 3	// Scene capable does not apply for central scene
											}
											var result = props.scene ? 2 : 1
											if (SCObj.HasCooperConfiguration) {
												return result // Cooper can handle both Scene and non-scene direct devices 
											}
											if (j == 0) {
												return result // The first entry in the list decides scene vs non-scene mode
											}
											if (props.basicSetOnly) { // The target device is not scene capable
												if (!enableNonSceneDirect && mode.prefix != "T") {
													return 0  // A vera scene is enabled. We need Scene Activate mode unless we can figure which button was pushed indirectly by reading the toggling indicator.
												}
											}
											return result;
				  	 					});
									var controllable = SceneController_GetDeviceProperties(SceneController_get_device_object(mode[j].device))
									if (((mode.sceneControllable && controllable.scene) || SCObj.HasCooperConfiguration || SCObj.SendsCentralScene) && controllable.multiLevel && mode.prefix != "H" && mode.prefix != "P" && mode.prefix != "E") {
										html +=  '    <input type="checkbox"'+((mode[j].level || mode[j].level == 0) && mode[j].level != 255 ?' checked':'')+
												      ' id="LevelSelect_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'_'+j+'"'+
										              ' style="min-width:10px;margin-left:10px;"'+
										              ' onChange="SceneController_SelectDirectDevice('+SCObj.Id+',\''+mode.prefix+'\','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',\''+j+'\',2)"'+
										              '><span/>\n'+
										         '    Level: \n'+
										         '    <input '+inputClass+' id="Level_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'_'+j+'" type="number"'+
										              ' value="'+((!(mode[j].level || mode[j].level == 0) || mode[j].level == 255)?"":mode[j].level)+'" style="width:40px;"'+
										              ' onChange="SceneController_SelectDirectDevice('+SCObj.Id+',\''+mode.prefix+'\','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',\''+j+'\',3)">\n';
									}
									if ((mode.sceneControllable && controllable.scene && controllable.multiLevel) || (SCObj.SendsCentralScene && controllable.multiLevelDuration))  {
										html +=  '    <input type="checkbox"'+((mode[j].dimmingDuration || mode[j].dimmingDuration == 0) && mode[j].dimmingDuration != 255 ?' checked':'')+
												      ' id="DimmingDurationSelect_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'_'+j+'"'+
										              ' style="min-width:10px;margin-left:10px;"'+
										              ' onChange="SceneController_SelectDirectDevice('+SCObj.Id+',\''+mode.prefix+'\','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',\''+j+'\',4)"'+
										              '><span/>\n'+
										         '    Duration: \n'+
										         '    <input '+inputClass+' id="DimmingDuration_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'_'+j+'" type="number"'+
										              ' value="'+((!(mode[j].dimmingDuration || mode[j].dimmingDuration == 0) || mode[j].dimmingDuration == 255)?"":mode[j].dimmingDuration)+'"'+
										              ' style="width:40px;"'+
										              ' onChange="SceneController_SelectDirectDevice('+SCObj.Id+',\''+mode.prefix+'\','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',\''+j+'\',5)">\n';
									}
								}
								html += '  </td></tr>\n'
							} // for all devices in mode
						    var anyOrAllNumDevices = mode.length + (mode.veraSceneSet ? mode.veraSceneSet.length : 0)
							/* Status LED Actions */
							if (SCObj.NumStatusLEDs > 0 && !mode.negative) {
								var ledAction = mode.LEDAction ? mode.LEDAction : sceneActionLine ? "N" : "-";
								var ledColor = mode.LEDColor ? mode.LEDColor : "W";
								var flashPeriod = mode.flashPeriod ? mode.flashPeriod : 5;
								html += ' <tr id="Screen_C_'+button+'_'+eventNum+'_'+state+'_B" '+rowStyle+'>\n'+
										'  <td/>\n'+
								        '  <td align="left" colspan=' + (numColumns-1) + '>\n';
								if (SCObj.NumButtons == 2 && button == 2 && eventNum == 1) {
									// Single tap LED mode of Off button is always slaved to On button mode
									var name = SCObj.StatusLEDActionNames[ledAction]
									if (!name) {
										name = SCObj.StatusLEDActionNames["-"]
									}
									html += '   ' + name +'\n';
								} else {
								    html += '   <select '+selectClass+' id="SetLEDAction_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'"'+
								                ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',2)">\n';
								    for (var k=0; k < SCObj.StatusLEDActions.length; ++k) {
								    	var ledActionItem = SCObj.StatusLEDActions.charAt(k);
								    	var ledActionName = SCObj.StatusLEDActionNames[ledActionItem];
								    	var show;
								    	// Actions: -NL7654321AR
								    	// Modes: N2345SA  
								    	if (sceneActionLine) {
								    		show = ledActionItem != "-" && ledActionItem != "L";
								    	} else if (eventNum == 1) {
								    		// show = ledActionItem == "-" || ledActionItem == "N";
								  			show = ledActionItem != "R";
								    	} else {
								    		show = ledActionItem != "N" && ledActionItem != "R";
								    	}
								    	if (show) {
								    		var temporary = false;
								    		if (timeoutEnable && ledActionItem != "-" && !sceneActionLine) {
								    			if (timeoutEvent == "N" || timeoutEvent == "1") {  // Normal/Single tap level
								    				temporary = eventNum != 1 || ledActionItem != "N"
								    			} else if (timeoutEvent >= "2" && timeoutEvent <= "9") { // Multi-tap device level
								    				temporary = ledActionItem != "L" || timeoutEvent != eventNum
								    			} else if (timeoutEvent == "S") { // Individual LED status
								    				temporary = ledActionItem < "1" || ledActionItem > "9"
								    			} else if (timeoutEvent == "A") { // All LED status
								    				temporary = ledActionItem != "A"
								    			}
								    		}
								    		if (temporary) {
								    			ledActionName = "Temporarily " + ledActionName.charAt(0).toLowerCase() + ledActionName.slice(1); 
								    		}
								    		html += '    <option value="'+ledActionItem+'"'+(ledAction == ledActionItem ? ' selected' : '')+'>'+ledActionName+'</option>\n';
								    	}
								    }
									html += '   </select>\n';
								}
								if (ledAction != "-" && ledAction != "R") {
								    html += '   to ' +
							                '   <select '+selectClass+' id="LEDColor_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'"'+
							                    ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',3)" style="width:140px;">\n';
						            var lastColor = SCObj.StatusLEDColors.length;
						            if (ledAction == "N") {
						            	lastColor = (lastColor+1) / 2; // No flashing colors in normal mode
						            }
						            for (var k = 0; k < lastColor; ++k) {
						            	var colorValue = SCObj.StatusLEDColors.charAt(k)
								    	html += '    <option value="'+colorValue+'"'+(colorValue == ledColor ? ' selected' : '')+'>'+
								    				  SceneController_StatusLEDColorNames[colorValue]+'</option>\n';
						            }
						            html += '   </select>\n';
						            if (ledColor >= "a" && ledColor <= "z") {
						            	html += 'flash rate:<input '+inputClass+' id="FlashPeriod_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'" type="number" min="0.1" max="25.5" step="0.1"'+
										              ' value="'+(flashPeriod/10)+'"'+
										              ' style="width:40px;"'+
							             		      ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',6)"/> seconds\n'
						            }
						       		// Any or all choice for Status LEDs
									if (anyOrAllNumDevices > 1 && (mode.prefix == "O" || (mode.prefix >= "2" && mode.prefix <= "9") || mode.prefix == "M" /* momentary used for multi-state */)) {
										if (ledAction == "L") {
											if (mode.allDevices != "Min" && mode.allDevices != "Max" && mode.allDevices != "Average" && mode.allDevices != "First") {
												mode.allDevices = "Average";
											} 
							             	html +='to the <select '+selectClass+' id="AnyOrAll_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'"'+
							             		   ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',4)">\n'+
										           '  <option value="Max"'     + (mode.allDevices=="Max"      ? ' selected' : '') + '>Maximum</option>\n'+
										           '  <option value="Min"'     + (mode.allDevices=="Min"      ? ' selected' : '') + '>Minimum</option>\n'+
										           '  <option value="Average"' + (mode.allDevices=="Average"  ? ' selected' : '') + '>Average</option>\n'+
										           '  <option value="First"'   + (mode.allDevices=="First"    ? ' selected' : '') + '>First</option>\n'+
										           ' </select>\n'+
										           ' device level\n';
										} else {
											if ((anyOrAllNumDevices == 2 && mode.allDevices == "Most") ||
												(mode.allDevices != "Any" && mode.allDevices != "All" && mode.allDevices != "Most")) {
												mode.allDevices = "Any";
											} 
							             	html +='if <select '+selectClass+' id="AnyOrAll_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'"'+
							             		   ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',4)"'+
							             		   ' style="w1dth:' + (anyOrAllNumDevices > 2 ? '100' : '70') + 'px;">\n'+
							             		   '  <option value="First"' + (mode.allDevices=="First"  ? ' selected' : '') + '>The first</option>\n'+
										           '  <option value="Any"' + (mode.allDevices=="Any"  ? ' selected' : '') + '>Any</option>\n';
											if (anyOrAllNumDevices > 2) {
												html += '  <option value="Most"'+ (mode.allDevices=="Most" ? ' selected' : '') + '>At least ' + Math.round(anyOrAllNumDevices / 2) + '</option>\n';
											}
										    html +=  	'  <option value="All"' + (mode.allDevices=="All"  ? ' selected' : '') + '>' + (anyOrAllNumDevices == 2 ? 'Both' : 'All ' + anyOrAllNumDevices) + '</option>\n'+
												    	'</select>\n'+
												    	(mode.allDevices=="First" ? ' device is' : ' devices are') + ' on\n'
										}
									}
								}
								html += '   .</td>\n'+
									    ' </tr>\n'+
									    ' <tr id="Screen_C_'+button+'_'+eventNum+'_'+state+'_C" '+rowStyle+'><td align="left" colspan='+numColumns+'><div style="height: 10px;"/></td></tr>\n';
							} // SCObj.NumStatusLEDs > 0
							/* Any or all choice for Indicator*/
							if (anyOrAllNumDevices > 1 && (SCObj.HasIndicator || SCObj.NumStatusLEDs > 0) && mode.prefix == "T") {
								if (anyOrAllNumDevices == 2 & mode.allDevices == "Most") {
									mode.allDevices = "Any"
								} 
								// If toggling more than one device, choose "Any" or "All"
								html += ' <tr id="Screen_C_'+button+'_'+eventNum+'_'+state+'_A" '+rowStyle+'>\n'+
								        '  <td/><td colspan=' + (numColumns-1) + ' align="left">'+
								        	SCObj.IndicatorHighlighting + ' ' + button + ' if \n';
				             	html += '<select '+selectClass+' id="AnyOrAll_'+peerId+'_'+curScreen+'_'+stateButton+'_'+eventCode+'"'+
				             		    ' onChange="SceneController_ChangeMode('+SCObj.Id+','+peerId+',\''+curScreen+'\','+stateButton+',\''+eventCode+'\',5)"'+
				             		    ' style="w1dth:' + (anyOrAllNumDevices > 2 ? '100' : '70') + 'px;">\n'+
							            '  <option value="First"' + (mode.allDevices=="First"  ? ' selected' : '') + '>The first</option>\n'+
							            '  <option value="Any"' + (mode.allDevices=="Any"  ? ' selected' : '') + '>Any</option>\n';
								if (anyOrAllNumDevices > 2) {
									html += '  <option value="Most"'+ (mode.allDevices=="Most" ? ' selected' : '') + '>At least ' + Math.round(anyOrAllNumDevices / 2) + '</option>\n';
								}
							    html += '  <option value="All"' + (mode.allDevices=="All"  ? ' selected' : '') + '>' + (anyOrAllNumDevices == 2 ? 'Both' : 'All ' + anyOrAllNumDevices) + '</option>\n'+
									    '</select>\n'+
										(mode.allDevices=="First" ? ' device is' : ' devices are') + ' on\n'
									    '</td></tr>';
							}

						} // ScreenType != T or button == 1 or button == 5
					} // for state
				} // for eventNum
			} // for button
			if (extraLines < 2) {
				html += ' <tr id="Screen_D">\n'+
				        '  <td>&nbsp;</td>'+
				        ' </tr>\n';
			}
		}
		if (screenType == "T") {
			var temperatureDevice = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, "TemperatureDevice_"+curScreen, 0);
			if (!temperatureDevice) {
				temperatureDevice = 0;
			}
			html += ' <tr if="Screen_E" '+rowStyle+'>\n'+
		            '  <td colspan=' + numColumns + ' align="left">Choose a Z-Wave thermostat or temperature sensor: '+
			  	   SceneController_DeviceMenu('TemperatureDevice_'+peerId+'_'+curScreen,
			  	 	temperatureDevice,
			  	 	'onChange="SceneController_SelectTemperatureDevice('+SCObj.Id+','+peerId+',\''+curScreen+'\')"' /* + ' style="w1dth:200px;"' */,
			  	 	"none", selectClass,
			  	 	function(obj) {return (obj.category_num == 5 || obj.category_num == 17) && SceneController_IsDirectZWaveObject(obj) ? 3 : 0;} )+
				    '   </td>\n'+
				    ' </tr>\n';
		}
		if (SCObj.HasMultipleScreens) {
			html += ' <tr id="Screen_F" '+rowStyle+'>\n'+
			        '  <td align="left" colspan='+numColumns+'>\n'+
				    '   <input type="checkbox"'+(timeoutEnable?' checked':'')+' id="TimeoutEnable_'+peerId+'_'+curScreen+'" style="min-width:'+(SceneController_IsUI7() ? 20 : 10)+'px;" onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', false)" /><span></span>\n'+
				    '   Switch to screen:\n'+
		            '   <select '+selectClass+' id="TimeoutScreen_'+peerId+'_'+curScreen+'" onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', true)" style="w1dth:100px;">\n'+
				         SceneController_ScreenMenu(SCObj, timeoutScreen, curScreen, "You cannot switch to the screen that you are alreayd on", lcdVersion)+
			        '   </select>\n'+
			        '   if no button pressed after\n'+
			        '   <input '+inputClass+' id="TimeoutSeconds_'+peerId+'_'+curScreen+'" type="text" value="'+timeoutSeconds+'" style="width:30px;" onChange="SceneController_ChangeTimeout('+SCObj.Id+','+peerId+',\''+curScreen+'\', true)" />\n'+
			        '   seconds, and then execute \n'+
				    '   <button type="button" '+buttonClass+' '+(timeoutEnable?'':'disabled ')+'style="min-width:10px;'+
				     	 (SceneController_FindScene(peerId, timeoutSceneNum, 5)?';color:orange;':'')+(timeoutEnable?'':'background-color:#AAAAAA;')+'"'+
				  		 (timeoutEnable?' title="Device trigger:&#xa;   '+deviceName+':&#xa;      Scene number '+timeoutSceneNum+' is activated"':'')+
				         ' onClick="SceneController_SetScene('+peerId+','+timeoutSceneNum+',5,1,\''+SceneController_ScreenName(SCObj, curScreen)+' Screen timeout\')">Scene</button>\n'+
				    '   </td>\n'+
				    ' </tr>\n';
		}
		html += '</table>\n';
		SceneController_set_panel_html(html);
		SceneController_Placeholder = 0;
	} catch(e){
		log_message("SceneController_screens error: " + e + " " + e.stack);
	}
}

function SceneController_ChangeTimeout(SCObj, peerId, curScreen, enable) {
	var timeoutEnableObj = document.getElementById("TimeoutEnable_"+peerId+"_"+curScreen);
	var timeoutScreenObj = document.getElementById( "TimeoutScreen_"+peerId+"_"+curScreen);
	var TimeoutSecondsObj = document.getElementById( "TimeoutSeconds_"+peerId+"_"+curScreen);
	var TimeoutEventObj = document.getElementById( "TimeoutEvent_"+peerId+"_"+curScreen);
	var timeoutEnable = timeoutEnableObj.checked;
	var timeoutScreen = timeoutScreenObj ? timeoutScreenObj.value : curScreen;
	var timeoutSeconds = TimeoutSecondsObj.value;
	var timeoutEvent = TimeoutEventObj ? TimeoutEventObj.value : "1"
	if (enable || timeoutEnable) {
		var seconds = Number(timeoutSeconds);
		if (SCObj.HasMultipleScreens) {
			if (seconds < 5 || seconds > 3600) {
				window.alert("Timeout must be between five seconds and one hour.");
				return;
			}
		} else {
			if (seconds < 1 || seconds > 3600) {
				window.alert("Timeout must be between one second and one hour.");
				return;
			}
		}
	}
	if (enable && !timeoutEnable) {
		timeoutEnable = true;
		timeoutEnableObj.checked = true;
	}
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetScreenTimeout",{ Screen:curScreen,
																				Enable:String(timeoutEnable),
																				TimeoutScreen:timeoutScreen,
																				TimeoutSeconds:timeoutSeconds,
																				TimeoutEvent:timeoutEvent});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEnable_"+curScreen, String(timeoutEnable), false);
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutScreen_"+curScreen, timeoutScreen, false);
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutSeconds_"+curScreen, timeoutSeconds, false);
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "TimeoutEvent_"+curScreen, timeoutEvent, false);
	SceneController_Screens(SCObj, peerId);
}

function SceneController_ChangeNumLines(SCObj, peerId, curScreen) {
	var numLinesObj=document.getElementById("NumLines_"+peerId+"_"+curScreen);
	var numLines = numLinesObj ? parseInt(numLinesObj.value) : 5;
	SceneController_send_action(peerId,SID_SCENECONTROLLER,"SetNumLines",{Screen:curScreen,Lines:numLines});
	SceneController_set_device_state(peerId, SID_SCENECONTROLLER, "NumLines_"+curScreen, numLines, false);
	SceneController_Screens(SCObj, peerId);
}


function SceneController_FinishScene(peerId, sceneID, doSave, isNew, originalSave, originalCancel, event) {
	document.getElementById("confirm_save_scene").onclick=originalSave;
	document.getElementById("confirm_cancel_scene").onclick=originalCancel;
	if (doSave) {
		originalSave(event);
    	has_changes('new scene');
	}
	else {
		originalCancel(event);
		if (isNew) {
        	jsonp.userdata_remove("scene",get_scene_obj(sceneID));
		}
	}
	jQuery(EVLCD1_SelectedMenuButton).click();

	cpanel.elemDOM.show();	/* Do this double open to avoid select height bug */
    cpanel.open('device', get_device_obj(peerId).device_type, peerId.toString(), SceneController_Scenes_Tab);
}

// Work around UI7's needless 32 character scene name limit.
// In UI5, there is a practical limit to what can be displayed in a tile.
function SceneController_CreateSceneName(devName, label, action, maxLength) {
	label = label.replace(/\\r/g, " ");
	label = label.replace(/-/g, "");
	devName = devName.trim();
	label = label.trim();
	while(1) {
		var name = 	devName + " " + label + action;
		if (maxLength == 0 || maxLength >= name.length) {
			return name;
		}
		if (devName.match(/.+\sController/)) {
			devName = devName.slice(0, -11);
			continue;
		}
		if (devName.match(/Evolve\s.+/)) {
			devName = devName.slice(7);
			continue;
		}
		var lastSpace = devName.lastIndexOf(" ");
		if (lastSpace > 0) {
			devName = devName.slice(0, lastSpace);
			continue;
		}
		lastSpace = label.lastIndexOf(" ");
		if (lastSpace > 0) {
			label = label.slice(0, lastSpace);
			continue;
		}
		var truncLength =  maxLength - (" " + label + action).length;
		if 	(truncLength > 0) {
			devName = devName.slice(0, truncLength);
			continue;
		}
		truncLength =  maxLength - action.length - (devName + " " + label).length;
		return (devName + " " + label).slice(0,truncLength) + action;
	}
}

// Activate is 
// -1 for any, 
// 0 for off toggle, 
// 1 for on toggle 
// 2 for tap, momendary.
// 3 for hold
// 4 for release
// 5 for timeout
function SceneController_FindScene(deviceID, sceneNum, activate) {
	var triggerTemplate = (activate == 0 || activate == 4) ? 2 : 1;
    var numScenes=jsonp.ud.scenes.length;
	var numFound = 0;
    var sceneObjs = [];
    for (var i = 0; i < numScenes; i++) {
        var sceneObj = jsonp.ud.scenes[i];
		if (sceneObj.triggers &&
			sceneObj.triggers.length > 0) {
			for (var j = 0; j < sceneObj.triggers.length; j++) {
				if (sceneObj.triggers[j].device == deviceID &&
					(sceneObj.triggers[j].template == triggerTemplate || activate < 0) &&
					 sceneObj.triggers[j].arguments[0].value == sceneNum) {
					sceneObjs[numFound++] = sceneObj;
					break;
				}
			}
		}
    }
	return numFound ? sceneObjs : null;
}

// Return an object who's keys are the device numbers affected by the given scene object regardless of delay
function SceneController_GetSceneSet(sceneObjs) {
	if (!sceneObjs || sceneObjs.length == 0) {
		return null;
	}
	var result = [];
	for (var h = 0; h < sceneObjs.length; ++h) {
		var sceneObj = sceneObjs[h];
		for (var i = 0; i < sceneObj.groups.length; ++i) {
			var group = sceneObj.groups[i];
			for (var j = 0; j < group.actions.length; ++j) {
				var levelSet = false;
				var action = group.actions[j];
				if (action.action == "SetLoadLevelTarget") {
					var args = action.arguments;
					for (var k = 0; k < args.length; ++k) {
						var argument = args[k];
						if (argument.name == "newLoadlevelTarget") {
							result.push({device: action.device,
										 level: parseInt(argument.value, 0)})
						    levelSet = true;
							break;
						}
					}
				}
				if (!levelSet) {
					result.push({device: action.device})
				}
			}
		}
	}
	// remove duplicates. Keep the one with the highest level or where level exists.
	return result.sort(function(a, b) {
		var cmp = a.device - b.device;
		if (cmp == 0) {
			return (b.level ? b.level : 0) - (a.level ? a.level : 0);
		}
  		return cmp;
	}).filter(function(item, pos, ary) {
        return !pos || item.device != ary[pos - 1].device;
    });
}

function SceneController_SceneSetsEqual(s1, s2) {
	if (!s1 || s1.length==0) {
		return !s2 || s2.length==0;
	}
	if (!s2 || s2.length != s1.length) {
		return false;
	}
	for (var i = 0; i < s1.length; ++ i) {
		if (s1[i].device != s2[i].device) {
			return false;
		}
		if (s1[i].level != s2[i].level) {
			return false;
		}
	}
	return true;
}

function SceneController_GetOrCreateNewScene(deviceID, sceneNum, activate, taps, label) {
    var sceneObjs = SceneController_FindScene(deviceID, sceneNum, activate)
	var sceneObj = sceneObjs ? sceneObjs[0] : null;
	if (sceneObj) {
		if (sceneObj.groups == undefined) {
			sceneObj.groups =  [{
				delay: 0,
				actions: []
			}];
		}
        return [false, sceneObj, sceneObj.name];
    }
	// The scene was not found. Create a new one
	var suffix = ""
	switch(activate) {
		case 0: suffix = " off"; break;
		case 1: suffix = " on" ; break;
		case 2: 
			switch(taps) {
				case 1:  suffix = " tap";               break;
				case 2:  suffix = " double tap";        break;
				default: suffix = " " + taps + " taps"; break;
			}
			break;
		case 3:
		case 4:
			switch(taps) {
				case 1:  suffix = " hold";                       break;
				case 2:  suffix = " tap-hold";                   break;
				case 3:  suffix = " tap-tap-hold";               break;
				default: suffix = " " + (taps-1) + " taps-hold"; break;
			} 
			if (activate == 4) {
				suffix += "-release";
			}
			break;
		case 5: suffix = " timeout"; break;
	}
    var sceneID = SceneController_IsUI7() ? application.newSceneId() % 1000000 : new_scene_id();
	var devObj = SceneController_IsUI7() ? application.getDeviceById(deviceID) : get_device_obj(deviceID);
    var sceneName = SceneController_CreateSceneName(devObj.name, label, suffix, SceneController_IsUI7() ? 32 : 22);
	var newSceneObj = {
    	id: sceneID,
    	name: sceneName,
    	room: devObj.room,
    	// notification_only: deviceID,
    	triggers:  [{
    		name:(sceneName+" Trigger"),
    		enabled: "1",
    		template: activate ? 1 : 2,
    		device: deviceID,
    		arguments: [{
				id: 1,
				value: sceneNum
    		}],
    	}],
		groups: [{
			delay: 0,
			actions: []
		}],
        modeStatus: "0"
	};
	if (SceneController_IsUI7()) {
		application.userDataAdd("scene", newSceneObj);
	} else {
		jsonp.userdata_add("scene",newSceneObj);
	}
    return [true, newSceneObj, sceneName];
}

// start_eidt_scene in cpanel_data.js is buggy in 1.5.622
function SceneController_start_edit_scene_UI5(sceneID){
    try{
        sceneToEditId = sceneID;

        // clone might not work in other browsers.Checking it is a must
        sceneBackup = clone(get_node_obj(jsonp.ud.scenes,sceneID));
        sceneReference = get_node_obj(jsonp.ud.scenes,sceneID);
        fStartSceneCreator(false);
    }catch(e){
        log_message("Edit SceneController_start_edit_scene_UI5 error: " + e + " " + e.stack);
    }
}

function SceneController_SetScene(peerId, sceneNum, activate, taps, label)
{
	if (SceneController_IsUI7()) {
	    var t = SceneController_GetOrCreateNewScene(peerId, sceneNum, activate, taps, label);
		var isNew = t[0];
		var sceneObj = t[1];
		var sceneName = t[2];
		myInterface.showCreateScenePage(sceneObj.id);
		myInterface.openCreateSceneStepTwo();
		if (typeof(sceneObj.groups) != "object" || sceneObj.groups.length < 1 || (sceneObj.groups.length == 1 && sceneObj.groups[0].delay == 0)) {
			// Jump directly to immediate actions unless there are already one or more delayed actions.
			myInterface.showCreateSceneSelectDevicesForAction(0);
		}
        var labelEdit = isNew ? 'Create new scene:' : 'Edit Scene:';
        jQuery(View.idForCreateSceneContainerStepsTitle()).html(labelEdit + " " + SceneController_EscapeHTML(sceneName));
		myInterface.SceneController_Save_new_scenes = myInterface.view_scenes;
		myInterface.view_scenes = function() {
			if ('SceneController_Save_new_scenes' in myInterface) {
				myInterface.view_scenes = myInterface.SceneController_Save_new_scenes;
				myInterface.SceneController_Save_new_scenes = undefined;
			}
			myInterface.openDeviceCpanel( application.getDeviceById(peerId), application.getDeviceTemplateById(peerId));
			jQuery("#device_cpanel_top_bottom_element_1").click();
		}
	} else {  // UI5
		//console.log("SceneController_SetScene peerId="+peerId+" sceneNum="+sceneNum+" activate="+activate+" label="+label);
		var saveButton = document.getElementById("confirm_save_scene");
		var cancelButton = document.getElementById("confirm_cancel_scene");
	    EVLCD1_SelectedMenuButton = jQuery("#buttons > .button.selected")[0];

		if (!saveButton || !cancelButton) {
			alert("Incompatible UI: Could not find save and cancel buttons.");
			return;
		}
	    var t = SceneController_GetOrCreateNewScene(peerId, sceneNum, activate, taps, label);
		var isNew = t[0];
		var sceneObj = t[1];
	    cpanel.close();
	    SceneController_start_edit_scene_UI5(sceneObj.id);
	    jQuery(".module_title").html("Define Actions for button labeled: " + SceneController_EscapeHTML(label));
		var originalSave   = saveButton.onclick;
		var originalCancel = cancelButton.onclick;
		saveButton.onclick   = function(event) {SceneController_FinishScene(peerId, sceneObj.id, true,  isNew, originalSave, originalCancel, event);};
		cancelButton.onclick = function(event) {SceneController_FinishScene(peerId, sceneObj.id, false, isNew, originalSave, originalCancel, event);};
	}
}

function SceneController_Copy(SCObj, deviceId) {
	var peerId
	if(SceneController_IsZWaveChild(deviceId)){
    	peerId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
	}
	else {
    	peerId = deviceId;
    	deviceId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
	}
	var curScreen = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, CURRENT_SCREEN, 0);
	if (!curScreen || typeof curScreen != "string") {
		curScreen = SCObj.DefaultScreen;
	}
	if (!CopySwap) {
		CopySwap = {
			"IsCopy": true,
		    "IsRow": true,
		    "SourceRowNum": 1,
		    "SourceScreen": curScreen,
			"SourceDevice": peerId,
		    "DestRowNum": 2,
		    "DestScreen": (curScreen == SCObj.DefaultScreen ? "C2" : SCObj.DefaultScreen),
			"DestDevice": peerId,
		    "CopyScenes": false
		};
	}
	else {
		if (CopySwap.SourceDevice != peerId && CopySwap.DestDevice != peerId) {
			var sameDevice = CopySwap.SourceDevice == CopySwap.DestDevice;
			CopySwap.SourceDevice = peerId;
			if (sameDevice) {
				CopySwap.DestDevice = peerId;
			}
		}
		var bothThisDevice = CopySwap.SourceDevice == CopySwap.DestDevice;
		if (CopySwap.SourceScreen != curScreen && CopySwap.DestScreen != curScreen) {
			var sameScreen = CopySwap.SourceScreen == CopySwap.DestScreen;
			CopySwap.SourceScreen = curScreen;
			if (bothThisDevice && sameScreen) {
			    CopySwap.DestScreen = curScreen;
			}
			else {
			    CopySwap.DestScreen = SceneController_OtherCompatibleScreen(CopySwap.DestScreen, curScreen);
			}
		}
	}
	SceneController_UpdateCopyTab(SCObj, deviceId)
}

// Function to draw the Copy lines/pages tab
function SceneController_UpdateCopyTab(SCObj, deviceId) {
	try {
		var peerId
		if(SceneController_IsZWaveChild(deviceId)){
	    	peerId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
		}
		else {
	    	peerId = deviceId;
	    	deviceId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER, PEER_ID, 0));
		}
		var versionInfo = SceneController_get_device_state(deviceId, ZWDEVICE_SID, VERSION_INFO, 0)
		var lcdVersion = versionInfo ? (parseInt(/,(\d+)$/.exec(versionInfo)[1])) : 39;
		var curScreen = SceneController_get_device_state(peerId, SID_SCENECONTROLLER, CURRENT_SCREEN, 0);
		if (!curScreen || typeof curScreen != "string") {
			curScreen = SCObj.DefaultScreen;
		}
		var screenType = curScreen.charAt(0);
		var screenNum = parseInt(curScreen.slice(1));
		var hasCustomLabels = (screenType == 'C' || (screenType == 'T' && screenNum > 2));
		var extraLines = 0;
		var html="Use this tab to copy or swap lines or whole screens betwen this or other LCD1's in your Z-Wave network.<br>"
		html += '  <select class="styled" align="left" style="w1dth:75px; height:22px;" id="CopySwap_'+peerId+'" onChange="SceneController_ChangeCopySwap('+peerId+')">';
			html += '   <option style="height:22px;" value="Copy" selected>Copy</option>';
			html += '   <option style="height:22px;" value="Swap"         >Swap</option>';
		html += '  </select> ';
		html += '  <select class="styled" align="left" style="w1dth:75px; height:22px;" id="RowPage_'+peerId+'" onChange="SceneController_ChangeRowPage('+peerId+')">';
			html += '   <option style="height:22px;" value="Row"  selected>Row</option>';
			html += '   <option style="height:22px;" value="Page"         >Page</option>';
		html += '  </select> ';
		SceneController_set_panel_html(html);
	} catch(e){
		log_message("SceneController_UpdateCopyTab error: " + e + " " + e.stack);
	}
}

var GenGenSceneController
if (SceneController_IsUI7()) {
  GenGenSceneController = (function (api) {
	return {
		uuid: 'fcdb51b0-36ea-45a1-ba59-191c77fe23db',

		// This is a nasty hack at several levels. We need to have the "Delete Device" button in the top level control panel
		// tab delete the Z-Wave device and not the peer device. In UI7, the "AfterInit" hook is called after the
		// delete device handler is established. We can replace it by using idForDeviceCpanelDeleteDeviceControl
		// but really we only want to change the device its operating on.
		// JQuery does not have a documented API for getting the current handlers of an element and the method to do so
		// changed across versions.
		// Unfortunately, getting the old handler is only half the problem. The click() function takes no
		// parameters and instead relies on the upvalue of the "device" variable in the closure. We need to re-evaluate
		// the handler with the correct Device (and "that" = the Interface object) in order to break lexical scope.
		// Just in case, we put this all in an try/catch as it may be fragile.
	 	AfterInit: function(SCObj, deviceId) {
			try {
				if (!SceneController_IsZWaveChild(deviceId)) {
				    var peerId=parseInt(SceneController_get_device_state(deviceId, SID_SCENECONTROLLER,PEER_ID, 0));
				    var ShowSingleDevice = SceneController_get_device_state(peerId, SID_SCENECONTROLLER,SHOW_SINGLE_DEVICE, 0);
					if (ShowSingleDevice != "0") {
						var device = SceneController_get_device_object(peerId)
						var that = api.ui
						var elem = jQuery(View.idForDeviceCpanelDeleteDeviceControl())
						var oldHandler = jQuery._data(elem[0], "events").click[0].handler
						eval("var newHandler="+oldHandler.toString())
						elem.off("click").on("click", newHandler)
					}
				}
		    } catch (e) {
		        Utils.logError("Error in GenGenSceneController.AfterInit: " + e + " " + e.stack);
		    }
		},

		EvolveLCD1_AfterInit: function(deviceId) {
			 this.AfterInit(EVOLVELCD1, deviceId);
		},

		CooperRFWC5_AfterInit: function(deviceId) {
			 this.AfterInit(COOPERRFWC5, deviceId);
		},

		NexiaOneTouch_AfterInit: function(deviceId) {
			 this.AfterInit(NEXIAONETOUCH, deviceId);
		},

		HomeSeerWS100_AfterInit: function(deviceId) {
			 this.AfterInit(HOMESEERWS100, deviceId);
		},

		HomeSeerWD100_AfterInit: function(deviceId) {
			 this.AfterInit(HOMESEERWD100, deviceId);
		},

		HomeSeerWS200_AfterInit: function(deviceId) {
			 this.AfterInit(HOMESEERWS200, deviceId);
		},

		HomeSeerWD200_AfterInit: function(deviceId) {
			 this.AfterInit(HOMESEERWD200, deviceId);
		},

		HomeSeerFC200_AfterInit: function(deviceId) {
			 this.AfterInit(HOMESEERFC200, deviceId);
		}
	}
  })(api);
}
//# sourceURL=J_GenGenSceneController.js
